"""This module contains classes for representing units that are part of a
given system of units.
"""

import enum
import string
from typing import Callable, List, Optional, Tuple, Union

import numpy as np

from pyxx.arrays.functions.equality import is_array_equal
from pyxx.strings.functions.content import str_includes_only
from pyxx.units.exceptions import (
    IncompatibleUnitsError,
    InvalidUnitMathError,
)
from .unitsystem import UnitSystem, UnitSystemSI


class ConstantUnitMathConventions(enum.Enum):
    """Options for defining what happens when a :py:class:`Unit` object is
    multiplied by a constant

    When defining units, there are often cases where one unit is simply a
    multiple of another; for instance, a meter is 100 times the size of a
    centimeter, and an inch is 2.54 times the size of a centimeter.  Rather
    than defining entirely new units, it would be helpful if there was a quick
    way to generate them from existing units.

    These options provide precisely this functionality.  However, there is a
    significant complication: what convention should we take?  Should
    multiplying a unit of *meters* by 1000 result in a unit 1000 times larger
    (a *kilometer*), or should it return a unit 1000 times smaller (a
    *millimeter*, since :math:`1\\ m * \\frac{1000\\ mm}{1\\ m} = 1000\\ mm`)?

    Due to this ambiguity, by default PyXX prohibits multiplication of
    :py:class:`Unit` objects and constants.  However, by setting the
    :py:attr:`Unit.CONSTANT_MATH_CONVENTION` attribute to one of the values
    described on this page,  it is possible to override this default and
    specify the desired math convention for multiplying :py:class:`Unit`
    objects by constants.

    Warnings
    --------
    This is an advanced option.  If you are uncertain about how it functions,
    leave it set to the default value (:py:attr:`DISABLE`) -- you won't lose
    any functionality of the PyXX package; you'll only lose a minor shortcut
    intended purely for convenience.

    Notes
    -----
    In general, it is best to set this option at the beginning of your code,
    as it can get confusing if you set it later in your code and as a result
    different parts of your code follow different conventions.

    To select a math convention for multiplying :py:class:`Unit` objects by
    constants, include code similar to:

    >>> from pyxx.units import Unit, ConstantUnitMathConventions
    >>> Unit.CONSTANT_MATH_CONVENTION = ConstantUnitMathConventions.DISABLE
    """
    #: Multiplying a unit by a constant is forbidden, and any attempt to do so
    #: results in an error being thrown.
    DISABLE = enum.auto()

    #: To create a new unit, multiply or divide by a constant as if you were
    #: directly manipulating the units themselves.
    #:
    #: For example, suppose that we have already defined a unit of
    #: milliseconds ``ms`` and we want to create a new unit seconds ``s``.
    #: Since one second is 1000 times larger than a millisecond, the unit of
    #: ``s`` could be created by:
    #:
    #: .. code-block:: python
    #:
    #:     s = 1000 * ms
    UNIT_BASED = enum.auto()

    #: To create a new unit, multiply or divide by a constant as you would do
    #: if converting quantities in the respective units.
    #:
    #: For example, suppose that we have already defined a unit of
    #: milliseconds ``ms`` and we want to create a new unit seconds ``s``.
    #: Since :math:`x` milliseconds is equal to :math:`0.001x` seconds, the
    #: unit of ``s`` could be created by:
    #:
    #: .. code-block:: python
    #:
    #:     s = 0.001 * ms
    QUANTITY_BASED = enum.auto()


class Unit:
    """Base class for representing a unit

    This class can be used to represent an arbitrary unit that is part
    of a given system of units.  The attributes of this class specify how
    the unit relates to the base units of the system of units.

    Examples
    --------
    For examples, refer to the :ref:`section-tutorials_units` page.
    """

    # Characters expected in unit names
    __unit_chars = string.ascii_letters + '_'

    #: Defines the mathematical conventions for multiplying or dividing a
    #: :py:class:`Unit` object by a constant
    #:
    #: Configures whether :py:class:`Unit` objects can be generated by
    #: multiplying or dividing another :py:class:`Unit` object by a constant,
    #: and if so what mathematical conventions are adopted.  For more detail,
    #: refer to the :py:class:`ConstantUnitMathConventions` documentation.
    #:
    #: The default is :py:attr:`ConstantUnitMathConventions.DISABLE`, which
    #: prohibits multiplying or dividing :py:class:`Unit` objects by constants.
    #:
    #: Warnings
    #: --------
    #: This is a class attribute, so changing its value for one class (such as
    #: :py:class:`UnitLinear`) will change it for ALL classes that are an
    #: instance of or inherit from :py:class:`Unit`.
    #:
    #: Notes
    #: -----
    #: To select a math convention for multiplying :py:class:`Unit` objects by
    #: constants, include code similar to:
    #:
    #: >>> from pyxx.units import Unit, ConstantUnitMathConventions
    #: >>> Unit.CONSTANT_MATH_CONVENTION = ConstantUnitMathConventions.DISABLE
    #:
    #: In general, it is best to set this option at the beginning of your code,
    #: as it can get confusing if you set it later in your code and as a result
    #: different parts of your code follow different conventions.
    CONSTANT_MATH_CONVENTION = ConstantUnitMathConventions.DISABLE

    def __init__(
            self,
            unit_system: UnitSystem,
            base_unit_exps: Union[List[float], Tuple[float, ...], np.ndarray],
            to_base_function: Callable[[np.ndarray, float], np.ndarray],
            from_base_function: Callable[[np.ndarray, float], np.ndarray],
            identifier: Optional[str] = None,
            name: Optional[str] = None):
        """Creates an instance of the :py:class:`Unit` class

        Defines an object representing a base or derived unit that is part of
        a given system of units.

        Parameters
        ----------
        unit_system : UnitSystem
            The system of units to which the unit belongs
        base_unit_exps : list or tuple or np.ndarray
            A 1D list of exponents relating the given object's unit to the
            base units of ``unit system``
        to_base_function : Callable
            A function that transforms a value in the given object's unit
            to the base units of ``unit_system``
        from_base_function : Callable
            A function that transforms a value in the base units of
            ``unit_system`` to the given object's unit
        identifier : str, optional
            A short identifier describing the unit (example: ``'kg'``)
            (default is ``None``)
        name : str, optional
            A name describing the unit (example: ``'kilogram'``) (default
            is ``None``)
        """
        # Store system of units
        if not isinstance(unit_system, UnitSystem):
            raise TypeError('Argument "unit_system" must be derived from '
                            '"pyxx.units.UnitSystem"')
        self._unit_system = unit_system

        # Store identifier and name
        if not (identifier is None or isinstance(identifier, str)):
            raise TypeError(
                'Argument "identifier" must be "None" or of type "str"')
        self._identifier = identifier

        if not (name is None or isinstance(name, str)):
            raise TypeError('Argument "name" must be "None" or of type "str"')
        self._name = name

        # Store exponents relating unit to base units
        self._base_unit_exps = np.array([float(i) for i in base_unit_exps])

        # Verify that number of base units matches that of unit system
        if len(self.base_unit_exps) != self.unit_system.num_base_units:
            raise ValueError(
                'Argument "base_unit_exps" implies '
                f'{len(self.base_unit_exps)} base units but unit '
                f'system "{self.unit_system}" has '
                f'{self.unit_system.num_base_units} base units')

        # Store functions to transform to or from base units
        self._to_base_function = to_base_function
        self._from_base_function = from_base_function

    def __mul__(self, value: Union['Unit', float]) -> 'Unit':
        # Verify that "value" is a valid unit or can be converted to
        # a constant unit
        if isinstance(value, Unit):
            unit = value
        elif self.CONSTANT_MATH_CONVENTION == ConstantUnitMathConventions.DISABLE:
            raise InvalidUnitMathError(
                'Unable to multiply units.  Constants can only be used as an '
                'operated if "CONSTANT_MATH_CONVENTION" attribute is NOT set '
                'to "DISABLE"')
        else:
            try:
                unit = self.__create_const_unit(value)
            except (TypeError, ValueError) as exception:
                raise InvalidUnitMathError(
                    f'Unable to multiply units.  Operand {value} is not of '
                    'type "Unit" or a constant') from exception

        # Set name and identifier of output unit
        if self.identifier is None or unit.identifier is None:
            identifier = None
        else:
            identifier = f'{self.identifier}*{unit.identifier}'

        if self.name is None or unit.name is None:
            name = None
        else:
            name = f'{self.name}*{unit.name}'

        # Output units after multiplication
        return Unit(
            unit_system        = self.unit_system,
            base_unit_exps     = self.base_unit_exps + unit.base_unit_exps,
            to_base_function   = lambda x, exp:
                self.to_base_function(unit.to_base_function(x, exp), exp),
            from_base_function = lambda x, exp:
                self.from_base_function(unit.from_base_function(x, exp), exp),
            identifier         = identifier,
            name               = name
        )

    def __pow__(self, exponent: float):
        # Ensure that exponent is a number
        exponent = float(exponent)
        exponent_str = f'{exponent}' if exponent >= 0 else f'({exponent})'

        # Set name and identifier of output unit
        if self.identifier is None:
            identifier = None
        elif str_includes_only(self.identifier, self.__unit_chars):
            identifier = f'{self.identifier}^{exponent_str}'
        else:
            identifier = f'({self.identifier})^{exponent_str}'

        if self.name is None:
            name = None
        elif str_includes_only(self.name, self.__unit_chars):
            name = f'{self.name}^{exponent_str}'
        else:
            name = f'({self.name})^{exponent_str}'

        return Unit(
            unit_system        = self.unit_system,
            base_unit_exps     = exponent * self.base_unit_exps,
            to_base_function   = lambda x, exp:
                self.to_base_function(x, exp*exponent),
            from_base_function = lambda x, exp:
                self.from_base_function(x, exp*exponent),
            identifier         = identifier,
            name               = name
        )

    def __truediv__(self, value: Union['Unit', float]) -> 'Unit':
        # Verify that "value" is a valid unit or can be converted to
        # a constant unit
        if isinstance(value, Unit):
            divisor = value
        elif self.CONSTANT_MATH_CONVENTION == ConstantUnitMathConventions.DISABLE:
            raise InvalidUnitMathError(
                'Unable to multiply units.  Constants can only be used as an '
                'operated if "CONSTANT_MATH_CONVENTION" attribute is NOT set '
                'to "DISABLE"')
        else:
            try:
                divisor = self.__create_const_unit(value)
            except (TypeError, ValueError) as exception:
                raise InvalidUnitMathError(
                    f'Unable to divide units.  Operand {value} is not of '
                    'type "Unit" or a constant') from exception

        # Set name and identifier of output unit
        if self.identifier is None or divisor.identifier is None:
            identifier = None
        elif str_includes_only(divisor.identifier, self.__unit_chars):
            identifier = f'{self.identifier}/{divisor.identifier}'
        else:
            identifier = f'{self.identifier}/({divisor.identifier})'

        if self.name is None or divisor.name is None:
            name = None
        elif str_includes_only(divisor.name, self.__unit_chars):
            name = f'{self.name}/{divisor.name}'
        else:
            name = f'{self.name}/({divisor.name})'

        # Output units after division
        return Unit(
            unit_system        = self.unit_system,
            base_unit_exps     = self.base_unit_exps - divisor.base_unit_exps,
            to_base_function   = lambda x, exp:
                self.to_base_function(divisor.to_base_function(x, -exp), exp),
            from_base_function = lambda x, exp:
                self.from_base_function(divisor.from_base_function(x, -exp), exp),
            identifier         = identifier,
            name               = name
        )

    def __repr__(self):
        return f'{self.__class__} {str(self)}'

    def __rmul__(self, value: Union['Unit', float]) -> 'Unit':
        return self.__mul__(value)

    def __rtruediv__(self, value: Union['Unit', float]) -> 'Unit':
        return (self.__truediv__(value))**(-1)

    def __str__(self):
        representation = ''

        if self.identifier is not None:
            representation += f'{self.identifier} - '

        if self.name is not None:
            representation += f'{self.name} - '

        representation += f'{self.base_unit_exps}'

        return representation

    def __create_const_unit(self, value: float):
        try:
            if self.CONSTANT_MATH_CONVENTION \
                    == ConstantUnitMathConventions.UNIT_BASED:
                return Unit(
                    unit_system        = self.unit_system,
                    base_unit_exps     = [0] * self.unit_system.num_base_units,
                    to_base_function   = lambda x, exp: (float(value))**exp * x,
                    from_base_function = lambda x, exp: x / (float(value))**exp,
                    identifier         = f'{float(value)}'
                )

            if self.CONSTANT_MATH_CONVENTION \
                    == ConstantUnitMathConventions.QUANTITY_BASED:
                return Unit(
                    unit_system        = self.unit_system,
                    base_unit_exps     = [0] * self.unit_system.num_base_units,
                    to_base_function   = lambda x, exp: x / (float(value))**exp,
                    from_base_function = lambda x, exp: (float(value))**exp * x,
                    identifier         = f'{float(value)}'
                )

        except (TypeError, ValueError) as exception:
            raise TypeError(
                f'Unable to create constant unit.  Constant {value} '
                'cannot be converted to type "float"') from exception

        # Ideally, this exception should never be thrown since caller
        # functions should check that "CONSTANT_MATH_CONVENTION" before
        # trying to create a constant unit.  However, it is still included
        # here as a backup
        raise InvalidUnitMathError(  # pragma: no cover
            'Unable to create constant unit.  Attribute '
            '"CONSTANT_MATH_CONVENTION" is not set to a recognized value')

    @property
    def base_unit_exps(self) -> np.ndarray:
        """A list of exponents relating the given object's units to the
        base units of :py:attr:`unit_system`"""
        return self._base_unit_exps

    @property
    def identifier(self) -> Union[str, None]:
        """A user-defined string that represents the unit
        (examples: kg, m, rad)"""
        return self._identifier

    @property
    def from_base_function(self):
        """A function that transforms a value in the base units of the system
        of units :py:attr:`unit_system` to the given object's units"""
        return self._from_base_function

    @property
    def name(self) -> Union[str, None]:
        """A user-defined string that describes the unit
        (examples: kilogram, meter, radian)"""
        return self._name

    @property
    def to_base_function(self):
        """A function that transforms a value from the given object's units
        to the base units of :py:attr:`unit_system`"""
        return self._to_base_function

    @property
    def unit_system(self) -> UnitSystem:
        """The system of units to which the unit belongs"""
        return self._unit_system

    def convert(self, value: Union[np.ndarray, list, tuple, float],
                convert_type: str, unit: 'Unit') -> np.ndarray:
        """Converts a quantity from one unit to another

        This method performs a unit conversion, converting one or more values
        from this object's units to another unit.  This conversion can be
        performed in "either direction" -- either from this object's units to
        another unit, or from another unit to this object's units.

        Parameters
        ----------
        value : np.ndarray or list or tuple or float
            Quantities to convert to a different unit
        convert_type : str
            Must be either ``'to'`` or ``'from'``.  Describes whether to
            convert ``value`` from this object's units to the units specified
            by ``unit``, or vice versa
        unit : Unit
            The unit to convert ``value`` to or from

        Returns
        -------
        np.ndarray
            NumPy array of the same shape as ``value`` containing the
            quantities after performing the specified unit conversion
        """
        # Validate input argument types
        if not isinstance(convert_type, str):
            raise TypeError('Argument "convert_type" must be of type "str"')

        if not isinstance(unit, Unit):
            raise TypeError(f'Argument "unit" must be of type {Unit}')

        # Verify that units can be converted
        if not self.is_convertible(unit):
            raise IncompatibleUnitsError(
                'Cannot perform unit conversion: units are not compatible')

        # Convert inputs to NumPy array
        inputs = np.array(value, dtype=np.float64)

        # Perform unit conversion
        if convert_type.lower() == 'from':
            return self.from_base(unit.to_base(inputs))

        if convert_type.lower() == 'to':
            return unit.from_base(self.to_base(inputs))

        raise ValueError('Argument "convert_type" must be exactly one of the '
                         'following: ("from", "to")')

    def is_convertible(self, unit: 'Unit') -> bool:
        """Checks whether a unit can be converted to another unit

        Checks two units can be converted between each other (i.e., whether
        they belong to the same system of units and have the same
        :py:attr:`base_unit_exps` relating them to the base units).

        Parameters
        ----------
        unit : Unit
            Another :py:class:`Unit` instance

        Returns
        -------
        bool
            Returns ``True`` if this unit instance and ``unit`` belong to the
            same system of units and have the same :py:attr:`base_unit_exps`
            attribute, and ``False`` otherwise
        """
        # This does not use `isinstance()` or else units with system of units
        # `UnitSystem` could be converted to any other system of units.
        # Instead, users should create a unique `UnitSystem` subclass,
        # corresponding their own conventions.  This prevents accidentally
        # converting units based on different conventions
        if not (type(self.unit_system) is type(unit.unit_system)):  # noqa: E721
            return False

        return is_array_equal(self.base_unit_exps, unit.base_unit_exps)

    def from_base(self, value: Union[np.ndarray, list, tuple, float],
                  exponent: float = 1.0) -> np.ndarray:
        """Converts a value or array from base units of the unit
        system to the given unit

        Parameters
        ----------
        value : np.ndarray or list or tuple or float
            Value(s) to convert to base units
        exponent : float, optional
            Exponent to which the unit is raised (default is 1.0)

        Returns
        -------
        np.ndarray
            NumPy array with the same shape as ``value`` containing the
            value(s) in ``value`` expressed in base units

        Notes
        -----
        Use the ``exponent`` argument to handle units which are raised to a
        power.  For instance, to convert square kilometers to base units
        (square meters), set ``exponent`` to 2.
        """
        inputs = np.array(value, dtype=np.float64)

        return self.from_base_function(inputs, exponent)

    def to_base(self, value: Union[np.ndarray, list, tuple, float],
                exponent: float = 1.0) -> np.ndarray:
        """Converts a value or array from the given unit to the base
        units of the unit system

        Parameters
        ----------
        value : np.ndarray or list or tuple or float
            Value(s) to convert from base units to the given unit
        exponent : float, optional
            Exponent to which the unit is raised (default is 1.0)

        Returns
        -------
        np.ndarray
            NumPy array with the same shape as ``value`` containing the
            value(s) in ``value`` converted from base units to the given
            unit

        Notes
        -----
        Use the ``exponent`` argument to handle units which are raised to a
        power.  For instance, to convert to cubic kilometers from cubic
        meters (the base unit), set ``exponent`` to 3.
        """
        inputs = np.array(value, dtype=np.float64)

        return self.to_base_function(inputs, exponent)


class UnitLinear(Unit):
    """Class for representing units with linear transformations to/from
    the base units

    Defines a unit in which the transformations to/from the base units
    of the system of units (that is, the functions given by
    :py:attr:`from_base_function` and :py:attr:`to_base_function`) are linear.
    A large portion of the units encountered in everyday use can be considered
    linear units, so this class was created to simplify defining such units.

    Notes
    -----
    To convert an array of arbitrary dimensions ``inputs`` from the given
    object's unit to the base units, the following equation is applied:

    .. code-block:: python

        outputs = (scale * inputs) + offset

    Examples
    --------
    For examples, refer to the :ref:`section-tutorials_units` page.
    """

    def __init__(
            self, unit_system: UnitSystem,
            base_unit_exps: Union[List[float], Tuple[float, ...], np.ndarray],
            scale: float,
            offset: float,
            identifier: Optional[str] = None,
            name: Optional[str] = None):
        """Creates an instance of the :py:class:`UnitLinear` class

        Defines an object representing a base or derived unit in which the
        functions converting a value to/from the base units of the system
        of units :py:attr:`unit_system` are linear functions.

        Parameters
        ----------
        unit_system : UnitSystem
            The system of units to which the unit belongs
        base_unit_exps : list or tuple or np.ndarray
            A 1D list of exponents relating the given object's unit to the
            base units of ``unit system``
        scale : float
            The multiplicative factor applied when converting from
            the given object's unit to the base units
        offset : float
            The constant value added when converting from the given
            object's unit to the base units
        identifier : str, optional
            A short identifier describing the unit (example: ``'kg'``)
            (default is ``None``)
        name : str, optional
            A name describing the unit (example: ``'kilogram'``) (default
            is ``None``)
        """
        # Store inputs
        if not isinstance(scale, (float, int, np.number)):
            raise TypeError('Argument "scale" must be of type "float"')
        self._scale = float(scale)

        if not isinstance(offset, (float, int, np.number)):
            raise TypeError('Argument "offset" must be of type "float"')
        self._offset = float(offset)

        # Initialize object
        super().__init__(
            unit_system        = unit_system,
            base_unit_exps     = base_unit_exps,
            to_base_function   = lambda x, exp: (scale**exp) * x + offset,
            from_base_function = lambda x, exp: (x - offset) / (scale**exp),
            identifier         = identifier,
            name               = name
        )

    def __str__(self):
        return f'{super().__str__()} - scale: {self.scale} - offset: {self.offset}'

    @property
    def offset(self) -> float:
        """The constant value added when converting from the given
        object's unit to the base units"""
        return self._offset

    @property
    def scale(self) -> float:
        """The multiplicative factor applied when converting from
        the given object's unit to the base units"""
        return self._scale


class UnitLinearSI(UnitLinear):
    """Class for representing units with linear transformations to/from
    the base units and using the International System of Units (SI)

    This class is very similar to :py:class:`UnitLinear`, with the
    addition that the system of units is automatically set to SI.

    Examples
    --------
    For examples, refer to the :ref:`section-tutorials_units` page.

    See Also
    --------
    UnitLinear :
        Nearly identical class, but additionally allows users to specify
        an arbitrary system of units
    """

    def __init__(
            self,
            base_unit_exps: Union[List[float], Tuple[float, ...], np.ndarray],
            scale: float,
            offset: float,
            identifier: Optional[str] = None,
            name: Optional[str] = None):
        """Creates an instance of the :py:class:`UnitLinearSI` class

        Defines an object representing a base or derived unit in which the
        functions converting a value to/from the base units of the system
        of units :py:attr:`unit_system` are linear functions, and the unit
        belongs to the SI system of units.

        Parameters
        ----------
        base_unit_exps : list or tuple or np.ndarray
            A 1D list of exponents relating the given object's unit to the
            base units of ``unit system``
        scale : float
            The multiplicative factor applied when converting from
            the given object's unit to the base units
        offset : float
            The constant value added when converting from the given
            object's unit to the base units
        identifier : str, optional
            A short identifier describing the unit (example: ``'kg'``)
            (default is ``None``)
        name : str, optional
            A name describing the unit (example: ``'kilogram'``) (default
            is ``None``)
        """
        super().__init__(
            unit_system    = UnitSystemSI(),
            base_unit_exps = base_unit_exps,
            scale          = scale,
            offset         = offset,
            identifier     = identifier,
            name           = name
        )



"""This module contains a set of commonly-used units.  All units are defined
in terms of the SI system (https://www.nist.gov/pml/owm/metric-si/si-units),
with the following sequence of seven base units:

1. Length: meter [:math:`m`]
2. Time: second [:math:`s`]
3. Amount of substance: mole [:math:`mol`]
4. Electric current: ampere [:math:`A`]
5. Temperature: Kelvin [:math:`K`]
6. Luminous intensity: candela [:math:`cd`]
7. Mass: kilogram [:math:`kg`]
"""

import math
from typing import Any, Dict

from .unit import UnitLinearSI


## PHYSICAL CONSTANTS --------------------------------------------------------
# Assumed physical constants used in performing unit conversions

# Atmospheric pressure
ATMOSPHERIC_PRESSURE_PA = 101325    # Pa (absolute pressure)

# Standard acceleration due to gravity
# https://physics.nist.gov/cgi-bin/cuu/Value?gn
STANDARD_GRAVITY = 9.80665  # m/s^2


## CONVERSION FACTORS --------------------------------------------------------
# Conversion factors used internally by the unit converter

# Inches to meters
_IN_TO_M_SCALE = 0.0254  # m/in

# Pound-mass to kilograms
_LBM_TO_KG_SCALE = 0.45359237  # kg/lbm

# Pound-force to Newtons
_LBF_TO_N_SCALE = 4.4482216152605  # N/lbf
# _LBF_TO_N_SCALE = _LBM_TO_KG_SCALE * STANDARD_GRAVITY

# Pounds per square inch (psi) to pascals (Pa)
_PSI_TO_PA_SCALE = 4.4482216152605 / 0.00064516  # Pa/psi
#                = _LBF_TO_N_SCALE / (_IN_TO_M_SCALE**2)

# Degrees Celsius to Kelvin
_DEG_C_TO_K_OFFSET = 273.15

# Degrees Fahrenheit to Kelvin
_DEG_F_TO_K_SCALE = 5/9
_DEG_F_TO_K_OFFSET = -160/9 + 273.15
#                  = -32 * _DEG_F_TO_K_SCALE + _DEG_C_TO_K_OFFSET

# Gallons to cubic meters
_GAL_TO_M3_SCALE = 0.003785411784  # gal/m^3


## UNIT CONVERTER DEFAULT UNITS ----------------------------------------------

SAMPLE_SI_UNITS: Dict[str, Dict[str, Any]] = {
    ## BASE UNITS ------------------------------------------------------------
    'm': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('length',),
        'name': 'meter',
        'aliases': ('meter', 'meters'),
    },
    's': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('time',),
        'name': 'second',
        'aliases': ('sec', 'second', 'seconds'),
    },
    'mol': {
        'unit': UnitLinearSI((0, 0, 1, 0, 0, 0, 0), scale=1, offset=0),
        'name': 'mole',
        'aliases': ('moles', 'mole'),
    },
    'A': {
        'unit': UnitLinearSI((0, 0, 0, 1, 0, 0, 0), scale=1, offset=0),
        'name': 'ampere',
        'description': 'Unit of measure of electric current',
        'aliases': ('amp', 'amps'),
    },
    'K': {
        'unit': UnitLinearSI((0, 0, 0, 0, 1, 0, 0), scale=1, offset=0),
        'tags': ('temperature',),
        'name': 'Kelvin',
        'description': 'Absolute temperature or temperature increment in Kelvin',
    },
    'cd': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 1, 0), scale=1, offset=0),
        'tags': ('luminance',),
        'name': 'candela',
    },
    'kg': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=1, offset=0),
        'tags': ('mass',),
        'name': 'kilogram',
        'description': 'Unit of measure of mass',
        'aliases': ('kilogram', 'kilograms'),
    },

    ## DIMENSIONLESS QUANTITIES ----------------------------------------------
    '-': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('dimensionless',),
        'name': 'dimensionless',
        'description': 'Unit assigned to dimensionless numbers',
        'aliases': ('dimensionless', 'unitless'),
    },

    ## LENGTH ----------------------------------------------------------------
    'mm': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('length',),
        'name': 'millimeter',
        'aliases': ('millimeter', 'millimeters'),
    },
    'cm': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=0.01, offset=0),
        'tags': ('length',),
        'name': 'centimeter',
        'aliases': ('centimeter', 'centimeters'),
    },
    'micron': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=1e-6, offset=0),
        'tags': ('length',),
        'name': 'micron',
        'aliases': ('microns', 'μm'),
    },
    'km': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=1000, offset=0),
        'tags': ('length',),
        'name': 'kilometer',
        'aliases': ('kilometer', 'kilometers'),
    },
    'in': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=_IN_TO_M_SCALE,
                             offset=0),
        'tags': ('length',),
        'name': 'inch',
        'aliases': ('inch', 'inches'),
    },
    'ft': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=0.3048, offset=0),
        'tags': ('length',),
        'name': 'foot',
        'aliases': ('foot', 'feet'),
    },
    'mi': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=1609.344, offset=0),
        'tags': ('length',),
        'name': 'mile',
        'aliases': ('mile', 'miles'),
    },
    'yd': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=0.9144, offset=0),
        'tags': ('length',),
        'name': 'yard',
        'aliases': ('yard', 'yards'),
    },
    'league': {
        'unit': UnitLinearSI((1, 0, 0, 0, 0, 0, 0), scale=4828.032, offset=0),
        'tags': ('length',),
        'name': 'league',
        'aliases': ('leagues',),
    },

    ## AREA ------------------------------------------------------------------
    'ac': {
        'unit': UnitLinearSI((2, 0, 0, 0, 0, 0, 0), scale=4046.8564224, offset=0),
        'tags': ('area',),
        'name': 'acre',
        'aliases': ('acre', 'acres'),
    },

    ## VOLUME ----------------------------------------------------------------
    'L': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('volume',),
        'name': 'liter',
        'aliases': ('liter', 'liters'),
    },
    'mL': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=1e-6, offset=0),
        'tags': ('volume',),
        'name': 'milliliter',
        'aliases': ('milliliter', 'milliliters'),
    },
    'gal': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE,
                             offset=0),
        'tags': ('volume',),
        'name': 'liquid gallon',
        'aliases': ('gallon', 'gallons'),
    },
    'qt': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE/4,
                             offset=0),
        'tags': ('volume',),
        'name': 'liquid quart',
        'aliases': ('quart', 'quarts'),
    },
    'pt': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE/8,
                             offset=0),
        'tags': ('volume',),
        'name': 'liquid pint',
        'aliases': ('pint', 'pints'),
    },
    'cup': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE/16,
                             offset=0),
        'tags': ('volume',),
        'name': 'liquid cup',
        'aliases': ('cups',),
    },
    'fl_oz': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE/128,
                             offset=0),
        'tags': ('volume',),
        'name': 'fluid ounce',
        'aliases': ('fluid_ounce',),
    },
    'tbsp': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE/256,
                             offset=0),
        'tags': ('volume',),
        'name': 'tablespoon',
        'aliases': ('tablespoon', 'tablespoons'),
    },
    'tsp': {
        'unit': UnitLinearSI((3, 0, 0, 0, 0, 0, 0), scale=_GAL_TO_M3_SCALE/768,
                             offset=0),
        'tags': ('volume',),
        'name': 'teaspoon',
        'aliases': ('teaspoon', 'teaspoons'),
    },

    ## TIME ------------------------------------------------------------------
    'ns': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=1e-9, offset=0),
        'tags': ('time',),
        'name': 'nanosecond',
        'aliases': ('nanosecond', 'nanoseconds'),
    },
    'microsec': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=1e-6, offset=0),
        'tags': ('time',),
        'name': 'microsecond',
        'aliases': ('microsecond', 'microseconds', 'μs'),
    },
    'ms': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=0.001, offset=0),
        'tags': ('time',),
        'name': 'millisecond',
        'aliases': ('millisecond', 'milliseconds'),
    },
    'min': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=60, offset=0),
        'tags': ('time',),
        'name': 'minute',
        'aliases': ('minute', 'minutes'),
    },
    'hr': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=3600, offset=0),
        'tags': ('time',),
        'name': 'hour',
        'aliases': ('hour', 'hours'),
    },
    'day': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=86400, offset=0),
        'tags': ('time',),
        'name': 'day',
        'aliases': ('days',),
    },
    'week': {
        'unit': UnitLinearSI((0, 1, 0, 0, 0, 0, 0), scale=604800, offset=0),
        'tags': ('time',),
        'name': 'week',
        'aliases': ('weeks',),
    },

    ## FREQUENCY -------------------------------------------------------------
    'Hz': {
        'unit': UnitLinearSI((0, -1, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('frequency',),
        'name': 'hertz',
        'aliases': ('hertz',),
    },
    'kHz': {
        'unit': UnitLinearSI((0, -1, 0, 0, 0, 0, 0), scale=1000, offset=0),
        'tags': ('frequency',),
        'name': 'kilohertz',
        'aliases': ('kilohertz',),
    },
    'MHz': {
        'unit': UnitLinearSI((0, -1, 0, 0, 0, 0, 0), scale=1e6, offset=0),
        'tags': ('frequency',),
        'name': 'megahertz',
        'aliases': ('megahertz',),
    },
    'GHz': {
        'unit': UnitLinearSI((0, -1, 0, 0, 0, 0, 0), scale=1e9, offset=0),
        'tags': ('frequency',),
        'name': 'gigahertz',
        'aliases': ('gigahertz',),
    },

    ## SPEED -----------------------------------------------------------------
    'mph': {
        'unit': UnitLinearSI((1, -1, 0, 0, 0, 0, 0), scale=1609.344/3600, offset=0),
        'tags': ('speed',),
        'name': 'miles per hour',
    },

    ## MASS ------------------------------------------------------------------
    'g': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=0.001, offset=0),
        'tags': ('mass',),
        'name': 'gram',
        'aliases': ('gram', 'grams'),
    },
    'mg': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=1e-6, offset=0),
        'tags': ('mass',),
        'name': 'milligram',
        'aliases': ('milligram', 'milligrams'),
    },
    'microgram': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=1e-9, offset=0),
        'tags': ('mass',),
        'name': 'microgram',
        'aliases': ('micrograms', 'μg'),
    },
    'lbm': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=_LBM_TO_KG_SCALE,
                             offset=0),
        'tags': ('mass',),
        'name': 'pound-mass',
        'description': ('Avoirdupois pound, as defined by NIST Handbook 44'),
        'aliases': ('pound-mass',),
    },
    't': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=1000, offset=0),
        'tags': ('mass',),
        'name': 'metric ton',
        'aliases': ('metric_ton',),
    },
    'tn': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=907.18474, offset=0),
        'tags': ('mass',),
        'name': 'short ton',
        'aliases': ('ton', 'tons', 'short_ton', 'short_tons'),
    },
    'long_ton': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=1016.0469088, offset=0),
        'tags': ('mass',),
        'name': 'long ton',
        'aliases': ('long_tons',),
    },
    'carat': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 1), scale=0.0002, offset=0),
        'tags': ('mass',),
        'name': 'carat',
        'aliases': ('carats',),
    },

    ## FORCE -----------------------------------------------------------------
    'N': {
        'unit': UnitLinearSI((1, -2, 0, 0, 0, 0, 1), scale=1, offset=0),
        'tags': ('force',),
        'name': 'Newtons',
        'aliases': ('Newton', 'Newtons'),
    },
    'kN': {
        'unit': UnitLinearSI((1, -2, 0, 0, 0, 0, 1), scale=1000, offset=0),
        'tags': ('force',),
        'name': 'kilonewtons',
        'aliases': ('kilonewton', 'kilonewtons'),
    },
    'lbf': {
        'unit': UnitLinearSI((1, -2, 0, 0, 0, 0, 1), scale=_LBF_TO_N_SCALE,
                             offset=0),
        'tags': ('force',),
        'name': 'pound-force',
        'description': ('Avoirdupois pound-force, as defined by NIST Handbook 44 '
                        'and using NIST\'s definition of standard acceleration '
                        f'due to gravity of {STANDARD_GRAVITY} m/s^2'),
        'aliases': ('pound-force',),
    },

    ## ANGLES ----------------------------------------------------------------
    'rad': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 0), scale=1, offset=0),
        'tags': ('angle',),
        'name': 'radian',
        'aliases': ('radian', 'radians'),
    },
    'deg': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 0), scale=math.pi/180, offset=0),
        'tags': ('angle',),
        'name': 'degrees',
        'aliases': ('degree', 'degrees'),
    },
    'rev': {
        'unit': UnitLinearSI((0, 0, 0, 0, 0, 0, 0), scale=2*math.pi, offset=0),
        'tags': ('angle',),
        'name': 'revolutions',
        'aliases': ('revolutions', 'rotations'),
    },

    ## PRESSURE --------------------------------------------------------------
    'Pa': {
        'unit': UnitLinearSI((-1, -2, 0, 0, 0, 0, 1), scale=1, offset=0),
        'tags': ('pressure',),
        'name': 'pascal',
        'description': 'Absolute pressure in units of pascals',
        'aliases': ('pascal', 'pascals'),
    },
    'kPa': {
        'unit': UnitLinearSI((-1, -2, 0, 0, 0, 0, 1), scale=1000, offset=0),
        'tags': ('pressure',),
        'name': 'kilopascal',
        'description': 'Absolute pressure in units of kilopascals',
        'aliases': ('kilopascal', 'kilopascals'),
    },
    'MPa': {
        'unit': UnitLinearSI((-1, -2, 0, 0, 0, 0, 1), scale=1e6, offset=0),
        'tags': ('pressure',),
        'name': 'megapascal',
        'description': 'Absolute pressure in units of megapascals',
        'aliases': ('megapascal', 'megapascals'),
    },
    'GPa': {
        'unit': UnitLinearSI((-1, -2, 0, 0, 0, 0, 1), scale=1e9, offset=0),
        'tags': ('pressure',),
        'name': 'gigapascal',
        'description': 'Absolute pressure in units of gigapascals',
        'aliases': ('gigapascal', 'gigapascals'),
    },
    'bar': {
        'unit': UnitLinearSI((-1, -2, 0, 0, 0, 0, 1), scale=1e5, offset=0),
        'tags': ('pressure',),
        'name': 'bar',
        'description': 'Absolute pressure in units of bar',
    },
    'psi': {
        'unit': UnitLinearSI((-1, -2, 0, 0, 0, 0, 1), scale=_PSI_TO_PA_SCALE,
                             offset=0),
        'tags': ('pressure',),
        'name': 'psi',
        'description': 'Absolute pressure in units of pounds per square inch',
    },

    ## TEMPERATURE -----------------------------------------------------------
    'degC': {
        'unit': UnitLinearSI((0, 0, 0, 0, 1, 0, 0), scale=1,
                             offset=_DEG_C_TO_K_OFFSET),
        'tags': ('temperature',),
        'name': 'Degrees Celsius',
        'description': 'Absolute temperature in degrees Celsius',
        'aliases': ('celsius', 'Celsius'),
    },
    'degF': {
        'unit': UnitLinearSI((0, 0, 0, 0, 1, 0, 0), scale=_DEG_F_TO_K_SCALE,
                             offset=_DEG_F_TO_K_OFFSET),
        'tags': ('temperature',),
        'name': 'Degrees Fahrenheit',
        'description': 'Absolute temperature in degrees Fahrenheit',
        'aliases': ('fahrenheit', 'Fahrenheit'),
    },
    'degR': {
        'unit': UnitLinearSI((0, 0, 0, 0, 1, 0, 0), scale=_DEG_F_TO_K_SCALE,
                             offset=0),
        'tags': ('temperature',),
        'name': 'Degrees Rankine',
        'description': ('Absolute temperature or temperature increment in '
                        'degrees Rankine'),
        'aliases': ('rankine', 'Rankine'),
    },
    'degC_diff': {
        'unit': UnitLinearSI((0, 0, 0, 0, 1, 0, 0), scale=1, offset=0),
        'tags': ('temperature',),
        'name': 'Degrees Celsius increment',
        'description': 'Temperature increment in degrees Celsius',
    },
    'degF_diff': {
        'unit': UnitLinearSI((0, 0, 0, 0, 1, 0, 0), scale=_DEG_F_TO_K_SCALE,
                             offset=0),
        'tags': ('temperature',),
        'name': 'Degrees Fahrenheit increment',
        'description': 'Temperature increment in degrees Fahrenheit',
    },

    # WORK/ENERGY ------------------------------------------------------------
    'J': {
        'unit': UnitLinearSI((2, -2, 0, 0, 0, 0, 1), scale=1, offset=0),
        'tags': ('work', 'energy',),
        'name': 'joule',
        'aliases': ('joule', 'joules'),
    },
    'mJ': {
        'unit': UnitLinearSI((2, -2, 0, 0, 0, 0, 1), scale=0.001, offset=0),
        'tags': ('work', 'energy',),
        'name': 'millijoule',
        'aliases': ('millijoule', 'millijoules'),
    },
    'kJ': {
        'unit': UnitLinearSI((2, -2, 0, 0, 0, 0, 1), scale=1000, offset=0),
        'tags': ('work', 'energy',),
        'name': 'kilojoule',
        'aliases': ('kilojoule', 'kilojoules'),
    },
    'MJ': {
        'unit': UnitLinearSI((2, -2, 0, 0, 0, 0, 1), scale=1e6, offset=0),
        'tags': ('work', 'energy',),
        'name': 'megajoule',
        'aliases': ('megajoule', 'megajoules'),
    },
    'GJ': {
        'unit': UnitLinearSI((2, -2, 0, 0, 0, 0, 1), scale=1e9, offset=0),
        'tags': ('work', 'energy',),
        'name': 'gigajoule',
        'aliases': ('gigajoule', 'gigajoules'),
    },

    # POWER ------------------------------------------------------------------
    'W': {
        'unit': UnitLinearSI((2, -3, 0, 0, 0, 0, 1), scale=1, offset=0),
        'tags': ('power',),
        'name': 'watt',
        'aliases': ('watt', 'watts'),
    },
    'mW': {
        'unit': UnitLinearSI((2, -3, 0, 0, 0, 0, 1), scale=0.001, offset=0),
        'tags': ('power',),
        'name': 'milliwatt',
        'aliases': ('milliwatt', 'milliwatts'),
    },
    'kW': {
        'unit': UnitLinearSI((2, -3, 0, 0, 0, 0, 1), scale=1000, offset=0),
        'tags': ('power',),
        'name': 'kilowatt',
        'aliases': ('kilowatt', 'kilowatts'),
    },
    'MW': {
        'unit': UnitLinearSI((2, -3, 0, 0, 0, 0, 1), scale=1e6, offset=0),
        'tags': ('power',),
        'name': 'megawatt',
        'aliases': ('megawatt', 'megawatts'),
    },
    'GW': {
        'unit': UnitLinearSI((2, -3, 0, 0, 0, 0, 1), scale=1e9, offset=0),
        'tags': ('power',),
        'name': 'gigawatt',
        'aliases': ('gigawatt', 'gigawatts'),
    },
}



"""**Functions and utilities to assist in debugging or analyzing Python code**

The :py:mod:`pyxx.dev` module is intended to provide commonly used tools
or scripts that can be useful when debugging Python code.
"""

from .classes import InterruptibleLoop, TimeIt



"""This module contains classes that provide nearly identical functionality
as built-in Python :py:class:`list` objects, but which require that all items
in a list be of a specified type
"""

from .typedlist import TypedList, TypedListWithID



"""Functions for converting data to an array, or converting one type of
array to another
"""

from typing import Any


def convert_to_tuple(value: Any) -> tuple:
    """Convert an input to a tuple

    Convert any input to a one-element tuple, or directly return
    the input (if it is already a tuple or list).

    Parameters
    ----------
    value : Any
        Value to convert to a tuple

    Returns
    -------
    tuple
        Directly returns ``value`` if it is already a tuple or list;
        otherwise, a one-element tuple containing ``value`` is returned

    Notes
    -----
    This function is intended to address an issue in which a value
    such as ``('value')`` is not considered a tuple (in this case,
    it would be a string), which may be unintuitive to users since
    it is in parentheses.  This function addresses this issue by
    converting any input to a tuple
    """
    return tuple(value) if isinstance(value, (tuple, list)) else (value,)



"""Functions for computing or analyzing file hashes
"""

import hashlib
import pathlib
from typing import Tuple, Union


def compute_file_hash(file: Union[str, pathlib.Path],
                      hash_function: str = 'sha256') -> Tuple[str, str]:
    """Computes a file hash

    Computes and returns the hash of a file on the disk. Users can select
    the hash function from those supported by ``hashlib``.

    Parameters
    ----------
    file : str or pathlib.Path
        File whose hash is to be computed
    hash_function : str, optional
        Hash function, selected from the options described in the
        "Notes" section (default is ``'sha256'``)

    Returns
    -------
    str
        Name of hash function (in the format used by ``hashlib``) which
        was computed
    str
        Hash of ``file`` as computed by the hash function specified
        by ``hash_function``

    Notes
    -----
    Any hash function included in Python's ``hashlib`` module can be used,
    as described here: https://docs.python.org/3/library/hashlib.html.

    Inputs for the ``hash_function`` argument are not case-sensitive, and
    dashes (``-``) and underscores (``_``) can be used interchangeably, or
    omitted.  Thus, to use the SHA-256 hash function, any of the following
    are valid inputs: ``'sha256'``, ``'SHA-256'``, ``'sha_256'``.
    """
    # Set hash function
    hash_name = hash_function.lower().replace('-', '').replace('_', '')
    file_hash_function = getattr(hashlib, hash_name)

    # Compute file hash
    file_hash: hashlib._Hash = file_hash_function()
    with open(file, 'rb') as fileID:
        while block := fileID.read(file_hash.block_size):
            file_hash.update(block)

    return (hash_name, file_hash.hexdigest())



"""**Code for processing lists, tuples, or other array-formatted data**

The :py:mod:`pyxx.arrays` module is intended to assist with processing data
stored in lists, tuples, NumPy arrays, or other matrix or tensor formats.
"""

from .classes import (
    TypedList,
    TypedListWithID,
)
from .functions import (
    convert_to_tuple,
    is_array_equal,
    check_len_equal,
    is_len_equal,
    max_list_item_len,
)



"""This module contains functions for parsing and analyzing strings
"""

from .content import (
    str_excludes_chars,
    str_includes_only,
)
from .brackets import (
    contains_all_matched_brackets,
    find_matching_bracket,
    find_skip_brackets,
    strip_matched_brackets,
)
from .split import split_at_index



"""Classes for timing code execution.
"""

import time

from pyxx.units.classes.unitconverter import UnitConverterSI


class TimeIt:
    """A context manager for measuring the duration of a code block.

    This context manager can be used in a "with" statement to measure the
    duration of code within the statement.  The resulting duration will be
    printed to the console when the "with" statement completes.

    Warnings
    --------
    Using this timer will add some overhead to the code block being measured.
    The measured duration outputted to the terminal will be slightly larger
    than the actual execution time of the code block.

    Examples
    --------

    Time a code block and print the duration to the terminal:

    .. code-block:: python

        >>> from pyxx.dev import TimeIt
        >>> import time
        >>> with TimeIt(units='ms', message='Execution time: {time:.2f} {units}'):
        ...     # Code block of which to measure the duration
        ...     time.sleep(1)
        Execution time: 1000.10 ms

    Time a code block and access the duration in Python code following the
    code block:

    .. code-block:: python

        >>> from pyxx.dev import TimeIt
        >>> import time
        >>> timer = TimeIt(print_duration=False)
        >>> with timer:
        ...     # Code block of which to measure the duration
        ...     time.sleep(1)
        >>> print(timer.duration('ms'))
        1000.1010101010101
    """

    def __init__(self, print_duration: bool = True, units: str = 's',
                 message: str = 'Code duration: {time} {units}') -> None:
        """Creates a new context manager for measuring the duration of a code block

        Parameters
        ----------
        print_duration : bool, optional
            Whether to print to the terminal the duration of the code block
        units : str, optional
            Only applicable if ``print_duration`` is ``True``.  Specifies the
            units in which the duration will be displayed to the terminal
            (default is ``'s'``)
        message : str, optional
            Only applicable if ``print_duration`` is ``True``.  The message
            template to display the duration (default is
            ``'Code duration: {time} {units}'``).  The ``{time}`` and
            ``{units}`` placeholders will be replaced by the duration and
            units, respectively
        """
        self.__unit_converter = UnitConverterSI()

        if (not (self.__unit_converter.is_defined_unit(units)
                 and self.__unit_converter.is_convertible('s', units))):
            raise ValueError(
                f'Invalid units: cannot convert from seconds to "{units}"')

        self.__print = bool(print_duration)
        self.__units = units
        self.__message = message

        self.__t0_s = 0.0
        self.__dt_s = 0.0

    def __enter__(self) -> None:
        self.__t0_s = time.time()

    def __exit__(self, *args, **kwargs) -> None:
        t1_s = time.time()

        # Calculate duration in seconds
        self.__dt_s = t1_s - self.__t0_s

        if self.__print:
            duration = self.__unit_converter.convert(
                quantity=self.__dt_s, from_unit='s', to_unit=self.__units)

            print(self.__message.format(time=duration, units=self.__units))

    def duration(self, units: str = 's') -> float:
        """Returns the last measured duration from the context manager

        Parameters
        ----------
        units : str
            The units in which to return the duration (default is ``'s'``)
        """
        return float(self.__unit_converter.convert(
            quantity=self.__dt_s, from_unit='s', to_unit=units
        ))



"""**Code for processing and managing numerical types of data**

The :py:mod:`pyxx.numbers` module contains a variety of code intended to aid
in processing numerical data.  For instance, this module contains code to
identify whether a value is a valid floating-point number or integer.
"""

from .functions import is_float, is_integer



"""Functions for splitting strings
"""


def split_at_index(value: str, index: int,
                   return_index: bool = False) -> tuple:
    """Splits a string at a given index

    This function can be used to split a string at a given position,
    returning the parts of the string before and after the given index, and
    optionally returning the character located at the index at which the
    string was split.

    Parameters
    ----------
    value : str
        String to split
    index : int
        Index specifying position at which to split ``value``
    return_index : bool, optional
        Whether to include the character located at ``index`` in the tuple of
        strings returned by the function (default is ``False``)

    Returns
    -------
    tuple
        A tuple containing the portions of ``value`` before and after
        the position specified by ``index``, and possibly the value
        ``value[index]`` (see Notes section for format details)

    Notes
    -----
    Suppose that ``begin``, ``delimiter``, and ``end`` are strings, and
    that ``value = f'{begin}{delimiter}{end}'``, where ``index = len(begin)``.

    - If ``return_index = False``, the following tuple is returned:
      ``(begin, end)``
    - If If ``return_index = True``, the following tuple is returned:
      ``(begin, delimiter, end)``
    """
    # Adjust if user provides an index relative to the end of the string
    if index < 0:
        index += len(value)

    if not return_index:
        return value[:index], value[index+1:]

    return value[:index], value[index], value[index+1:]



"""Collection of utilities for Python code"""

# PROGRAM VERSION ------------------------------------------------------------
__version__ = '1.1.0'


# PACKAGE MODULES ------------------------------------------------------------
from . import arrays
from . import dev
from . import files
from . import numbers
from . import strings
from . import units



"""Base class for processing files of any type (text or binary)
"""

import copy
import pathlib
from typing import Dict, Optional, Union

from pyxx.arrays.functions.convert import convert_to_tuple
from pyxx.files.functions.hash import compute_file_hash
from pyxx.files.exceptions import NoFileSpecifiedError, UntrackedFileError


class File:
    """Base class for processing files of any type (text or binary)

    This class is intended to represent an arbitrary file (which can but does
    not necessarily exist).  After creating a new instance of this class, it
    is possible to perform operations such as calculating file hashes and
    tracking whether the file has been modified.
    """

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None):
        """Define an arbitrary file

        Creates an object that represents and can be used to process a file of
        any type (text or binary).

        Parameters
        ----------
        path : str or pathlib.Path
            Path describing the location in the file system of the file that
            the object is to represent
        """
        # Initialize dictionary to store file hashes
        self._hashes: Dict[str, str] = {}

        # Store file path
        self.path = path  # type: ignore
                          # (workaround for python/mypy#3004)  # noqa: E114, E116

    def __repr__(self):
        # Display class
        representation = f'{__class__}\n'

        # Display path
        if self.path is not None:
            representation += f'--> File path: {str(self.path)}\n'

        # Display file hashes
        if len(self._hashes) > 0:
            header = '--> File hash:' \
                if len(self._hashes) == 1 else '--> File hashes:'
            representation += self.__file_hash_str(header, 4)

        return representation[:-1]

    def __str__(self):
        return f'{__class__} path="{self.path}"'

    def __file_hash_str(self, header: str, indent: int = 0):
        file_hash_str = f'{header}\n'

        for i in self._hashes.items():
            file_hash_str += f'{" " * indent}{i[0]}: {i[1]}\n'

        return file_hash_str

    @property
    def hashes(self) -> Dict[str, str]:
        """A copy of the dictionary containing any file hashes previously
        computed for the file specified by the :py:attr:`path` attribute"""
        return copy.deepcopy(self._hashes)

    @property
    def path(self) -> Union[pathlib.Path, None]:
        """Path describing the location of the file on the disk

        Assigning a value to this attribute (regardless whether it matches the
        current value or is a different path) will save the value as a
        ``pathlib.Path`` and **will automatically clear any saved file
        hashes**.
        """
        return self._path

    @path.setter
    def path(self, path: Optional[Union[str, pathlib.Path]]):
        # Clear any existing file hashes
        self.clear_file_hashes()

        # Store file path
        self._path = None if path is None else pathlib.Path(path)

    def clear_file_hashes(self) -> None:
        """Clears any stored file hashes"""
        self._hashes.clear()

    def compute_file_hashes(self,
            hash_functions: Union[tuple, str] = ('md5', 'sha256'),  # noqa : E128
            store: bool = False) -> Dict[str, str]:                 # noqa : E128
        """Computes hashes of the file specified by the :py:attr:`path`
        attribute

        Computes and returns the hashes of the file specified by the
        :py:attr:`path` attribute, with the option to populate the
        :py:attr:`hashes` dictionary with their values.

        Parameters
        ----------
        hash_functions : tuple or str, optional
            Tuple of strings (or individual string) specifying which hash(es)
            to compute. Any hash functions supported by ``hashlib`` can be
            used. Default is ``('md5', 'sha256')``
        store : bool, optional
            Whether to store the computed hashes in the :py:attr:`hashes`
            dictionary (default is ``False``)

        Returns
        -------
        dict
            A dictionary containing the file hashes specified by
            ``hash_functions``

        See Also
        --------
        pyxx.files.compute_file_hash :
            Function used to compute file hashes

        Notes
        -----
        Prior to calling this method, the :py:attr:`path` attribute must be
        defined.  To simultaneously set the :py:attr:`path` attribute and store
        file hashes, use :py:meth:`track_new_file`.
        """
        # Check that `path` attribute is defined
        if self.path is None:
            raise NoFileSpecifiedError(
                'Attribute "path" must be defined to compute file hashes')

        # Check that file exists
        if not self.path.exists():
            raise FileNotFoundError(
                f'Cannot compute hash for non-existent file "{self.path}"')

        if not self.path.is_file():
            raise IsADirectoryError(
                f'Cannot compute hash for a directory ("{self.path}")')

        # Ensure that inputs such as `hash_functions=('md5')` are still
        # interpreted as a tuple, not a string
        hash_functions = convert_to_tuple(hash_functions)

        # Compute file hash(es)
        output = {}
        for func in hash_functions:
            hash_name, file_hash = compute_file_hash(self.path, func)
            output[hash_name] = file_hash

            if store:
                self._hashes[hash_name] = file_hash

        return output

    def has_changed(self) -> bool:
        """Returns whether the file specified by the :py:attr:`path`
        attribute has changed since the last time file hashes were computed

        Returns
        -------
        bool
            Whether file has changed since the last time file hashes
            were computed
        """
        if len(self._hashes) == 0:
            raise UntrackedFileError(
                'File hashes have not yet been computed. Cannot '
                'evaluate whether file has changed')

        # Compute hashes of current file
        current_hashes = self.compute_file_hashes(tuple(self._hashes.keys()),
                                                  store=False)

        # Check whether current file hashes match those stored
        for key, value in self._hashes.items():
            if current_hashes[key] != value:
                return True

        return False

    def set_read_metadata(self,
                          path: Optional[Union[str, pathlib.Path]] = None
                          ) -> None:
        """Configures metadata related to file to be read from disk

        This method performs several pre-processing steps to prepare to read
        a file from the disk:

        1. Sets the :py:attr:`path` attribute.  If the ``path`` argument was
           provided, the attribute is set to this value; otherwise, the
           existing value stored in the :py:attr:`path` attribute is used (or
           an error is thrown if not defined).
        2. Verifies that the file specified by the :py:attr:`path` attribute
           exists.
        3. Stores the hashes for the file.

        It is advised that this method be called prior to reading any file.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            Location of the file in the file system (default is ``None``)

        Raises
        ------
        AttributeError
            If the both the ``path`` argument and the existing :py:attr:`path`
            attribute are ``None``
        FileNotFoundError
            If the file specified by :py:attr:`path` (after completing Step 1
            above) does not exist
        """
        # Identify file to read, and store in "path" attribute
        if path is not None:
            # First priority: use "path" argument if provided
            #   Mypy exclusion is a workaround for python/mypy#3004
            self.path = path  # type: ignore
        else:
            # Second priority: use "path" attribute.  If it isn't yet
            # defined, throw an error
            if self.path is None:
                raise AttributeError(
                    'Neither the "path" argument was provided nor is the '
                    '"path" attribute defined.  At least one of these must '
                    'be provided to read the file')

        # Check that file exists
        #   Mypy exclusion is a workaround for python/mypy#3004
        if not self.path.is_file():  # type: ignore
            raise FileNotFoundError(f'Cannot find file "{self.path}"')

        # Compute and store file hashes
        self.store_file_hashes()

    def store_file_hashes(self,
            hash_functions: Union[tuple, str] = ('md5', 'sha256')  # noqa : E128
            ) -> None:                                             # noqa : E128
        """Computes and stores hashes of the file specified by the
        :py:attr:`path` attribute

        Computes given hashes of the file specified by the :py:attr:`path`
        attribute and populates the :py:attr:`hashes` dictionary with their
        values.

        Parameters
        ----------
        hash_functions : tuple or str, optional
            Tuple of strings (or individual string) specifying which hash(es)
            to compute. Any hash functions supported by ``hashlib`` can be
            used. Default is ``('md5', 'sha256')``

        See Also
        --------
        pyxx.files.compute_file_hash :
            Function used to compute file hashes
        track_new_file :
            Use this method if you want to store file hashes but the
            :py:attr:`path` attribute isn't yet defined

        Notes
        -----
        Prior to calling this method, the :py:attr:`path` attribute must be
        defined.  To simultaneously set the :py:attr:`path` attribute and store
        file hashes, use :py:meth:`track_new_file`.
        """
        _ = self.compute_file_hashes(hash_functions, store=True)

    def track_new_file(self, path: Union[str, pathlib.Path],
            hash_functions: Union[tuple, str] = ('md5', 'sha256')  # noqa : E128
            ) -> None:                                             # noqa : E128
        """Shortcut for simultaneously modifying the :py:attr:`path` attribute
        and storing file hashes

        This method functions as a "shortcut," both modifying the
        :py:attr:`path` attribute and storing an optionally user-specified
        list of file hashes in the :py:attr:`hashes` attribute.  The
        intention of this method is that if a :py:class:`File` instance is
        tracking a given file, and user wants to switch to tracking another
        file, this provides a convenient way to do so with a single line
        of code.

        Parameters
        ----------
        file : str or pathlib.Path
            File that the object is to represent
        hash_functions : tuple or str, optional
            Tuple of strings (or individual string) specifying which hash(es)
            to compute. Any hash functions supported by ``hashlib`` can be
            used. Default is ``('md5', 'sha256')``

        See Also
        --------
        pyxx.files.compute_file_hash :
            Function used to compute file hashes
        """
        # Store the file path and name
        if path is None:
            raise TypeError('Argument "path" cannot be "None" when calling '
                            '"track_file()" method')

        self.path = path  # type: ignore
                          # (workaround for python/mypy#3004)  # noqa: E114, E116

        # Compute and store file hashes
        self.store_file_hashes(hash_functions)



"""Customized exceptions for the :py:mod:`pyxx.units` module
"""

## UNIT-PARSING ERRORS -------------------------------------------------------
# Errors thrown when issues are encountered when parsing strings
# containing units
class InvalidUnitError(ValueError):
    """Error thrown if a unit string is not formatted in a valid way"""

class InvalidUnitMathError(TypeError):
    """Error thrown if attempting to perform invalid math operations with
    :py:class:`Unit` objects"""

class ParserMaxIterationError(InvalidUnitError):
    """Error thrown if unit parser reaches its iteration limit"""


## UNIT ARITHMETIC ERRORS ----------------------------------------------------
# Errors thrown when attempting to perform invalid unit arithmetic operations
class InvalidExponentError(InvalidUnitError):
    """Error thrown if attempting to parse a unit string and a value is
    raised to a non-numeric exponent"""


## UNIT CONVERSION ERRORS ----------------------------------------------------
# Errors thrown when issues are encountered when attempting to perform
# unit conversions

# General unit conversion errors
class UnitConversionError(Exception):
    """General error for issues when performing unit conversions"""

class IncompatibleUnitsError(UnitConversionError):
    """Error thrown when attempting to convert a quantity between units, but
    the units are not compatible"""


# Errors specific to `UnitConverter` objects
class UnitConverterError(Exception):
    """General error for runtime issues related to :py:class:`UnitConverter`
    objects"""

class InvalidSearchFieldError(UnitConverterError, ValueError):
    """Error thrown when providing invalid search field(s) to the
    :py:meth:`pyxx.units.UnitConverter.search` method"""

class UnitNotFoundError(UnitConverterError, KeyError):
    """Error thrown when unable to retrieve a unit from a
    :py:class:`pyxx.units.UnitConverter` instance"""

class UnitAlreadyDefinedError(UnitConverterError):
    """Error thrown if attempting to add a unit to a :py:class:`UnitConverter`
    and the unit already exists"""



"""Functions for evaluating equality of array-like objects
"""

from __future__ import annotations
from numbers import Number
from typing import List, Tuple, Union

import numpy as np

from .size import is_len_equal


# Type alias: type for each `item` argument in `is_array_equal()`
__element_types = Union[Number, np.ndarray, str]
Array_or_Number_or_String = Union[List[__element_types],
                                  Tuple[__element_types],
                                  __element_types]


def is_array_equal(item1: Array_or_Number_or_String,
                   item2: Array_or_Number_or_String,
                   *args: Array_or_Number_or_String,
                   tol: float = 1e-16) -> bool:
    """Checks that arrays are equal in shape and content

    Returns ``True`` if all arrays passed as arguments are of the same shape
    and all elements are equal within a given tolerance ``tol`` (for numeric
    elements) or exactly equal (for string elements), and returns ``False``
    otherwise.  Inputs can be lists, tuples, NumPy arrays, numbers, strings,
    or nested arrays composed of any of these types.

    Parameters
    ----------
    item1 : list or tuple or np.ndarray or Number or str
        First array to evaluate
    item2 : list or tuple or np.ndarray or Number or str
        Second item to evaluate
    *args : list or tuple or np.ndarray or Number or str, optional
        Any other arrays to be evaluated
    tol : float, optional
        Maximum difference between numeric values to consider equivalent
        (default is ``1e-16``)

    Returns
    -------
    bool
        Whether ``item1``, ``item2``, ``*args`` have the same shape, and
        all elements are equal within tolerance ``tol`` (for numeric elements)
        and exactly equal (for string elements)

    Warnings
    --------
    - The shape of the input arrays must be identical for the arrays to be
      considered equal.  The shape of numbers is considered different from the
      shape of lists, so observe that ``0`` and ``[0]`` are **not** considered
      equal in shape.

    - By default, NumPy arrays are of homogeneous type.  This means that, for
      instance, ``pyxx.arrays.is_array_equal(np.array([1, 'a']), [1, 'a'])``
      evaluates to ``False`` (because the NumPy array is converted to all
      strings).  To avoid this issue, it is possible to create NumPy arrays
      with the ``dtype=object`` argument and allow mixed types.  For example,
      ``pyxx.arrays.is_array_equal(np.array([1, 'a'], dtype=object), [1, 'a'])``
      evaluates to ``True``.

    Notes
    -----
    **Recursion Limit**

    Internally, :py:func:`is_array_equal` is a recursive function.  It is
    possible that for extremely large nested arrays, Python's recursion limit
    may be reached.  If this occurs and it is necessary to compare such a
    large array, consider increasing the recursion limit using
    the `sys.setrecursionlimit() <https://docs.python.org/3/library/sys.html
    #sys.setrecursionlimit>`__ function.

    **Purpose**

    One question that may arise is, *why is this function necessary?*  NumPy
    already offers functions like `numpy.array_equal() <https://numpy.org/doc
    /stable/reference/generated/numpy.array_equal.html>`__, `numpy.isclose()
    <https://numpy.org/doc/stable/reference/generated/numpy.isclose.html>`__,
    and `numpy.allclose() <https://numpy.org/doc/stable/reference/generated
    /numpy.allclose.html>`__.

    There are several main advantages of :py:func:`is_array_equal`:

    - NumPy requires that arrays are numeric and are not "ragged" (sub-lists
      must all have the same length, recursively. For example, the array
      ``x = [[1,2,3], [1,2]]`` is "ragged" since ``len(x[0]) != len(x[1])``).
      In contrast, :py:func:`is_array_equal` can compare arrays with a mix of
      strings, numbers, lists, and tuples, as well as "ragged" arrays.

    - The NumPy functions mentioned will typically throw an exception if the
      array sizes being compared differ, but :py:func:`is_array_equal` simply
      returns ``False`` in this case.  This can eliminate the need to catch
      exceptions for certain applications.
    """
    # Create list of array(s) to compare with `item1`
    items = [item2] + list(args)

    # Check whether each of the input arguments is an array-like object
    is_array = [isinstance(x, (list, tuple, np.ndarray))
                for x in [item1, *items]]

    # If inputs are numbers, directly compare them (requiring difference
    # between numbers to be less than or equal to `tol` to consider the
    # inputs equal)
    if isinstance(item1, Number) \
            or (isinstance(item1, np.ndarray) and item1.ndim == 0):
        for x in items:
            # Check whether `item2` or any of `args` are an array.  If so,
            # this indicates that the array shapes are not equal
            if isinstance(x, (list, tuple)) \
                    or (isinstance(x, np.ndarray) and x.ndim > 0):
                return False

            # Argument `item1` is known to be a number, so attempt to see
            # whether each corresponding element in the other input arrays
            # is within tolerance `tol`
            try:
                # Disable Mypy warnings on the following line, since errors
                # will be handled with the try statement
                if abs(x - item1) > tol:  # type: ignore
                    return False

            except TypeError:
                return False

        return True

    # If inputs are array-like objects, compare their contents
    if any(is_array):
        # Verify that inputs are array-like objects
        if not all(is_array):
            return False

        # Verify that all inputs have equal length
        if not is_len_equal(item1, *items):
            return False

        # Check whether each sub-array's elements are equal (recursively)
        for i, x in enumerate(item1):

            # Disable Mypy warnings on the following line, since we've
            # already checked that the lengths of `x` and all elements
            # in `items` are equal
            if not is_array_equal(x, *[item[i] for item in items],  # type: ignore
                                  tol=tol):
                return False

        return True

    # Inputs are not numbers or array-like objects, so try to directly
    # compare them.  This allows strings, user-defined classes/types,
    # or other objects to be compared
    for x in items:
        if item1 != x:
            return False
    return True



"""Functions for evaluating or comparing sizes of array-like objects
"""

from typing import Any, Iterable


def check_len_equal(item1: Any, item2: Any, *args: Any):
    """Checks whether the lengths of a set of sequence-type objects
    are equal

    Evaluates the lengths of a set of items (such as lists, tuples, or
    strings), returning whether all items have the same length as well as
    either the length of all items (if all lengths are equal) or a list
    containing the lengths of each item (if they are not equal).

    Parameters
    ----------
    item1 : Any
        First item whose length to evaluate
    item2 : Any
        Second item whose length to evaluate
    *args : Any, optional
        Any other items whose lengths are to be evaluated

    Returns
    -------
    bool
        Whether all items have the same length
    int or list
        Returns an integer containing the length of all items (if all
        lengths are equal), or a list containing the lengths of each
        item (if lengths differ)

    See Also
    --------
    is_len_equal :
        Identical functionality, but returns only the ``bool`` output and
        may theoretically run slightly faster in cases where the length(s)
        of the inputs does not need to be returned
    """
    lengths = [len(item1)] + [len(item2)] + [len(i) for i in args]

    if len(set(lengths)) == 1:
        return True, lengths[0]

    return False, lengths


def is_len_equal(item1: Any, item2: Any, *args: Any) -> bool:
    """Checks whether the lengths of a set of sequence-type objects
    are equal

    Evaluates the lengths of a set of items (such as lists, tuples, or
    strings), returning whether all items have the same length.  This
    function should be slightly faster than :py:func:`check_len_equal`
    for applications where the lengths of the input arguments do not
    need to be returned.

    Parameters
    ----------
    item1 : Any
        First item whose length to evaluate
    item2 : Any
        Second item whose length to evaluate
    *args : Any, optional
        Any other items whose lengths are to be evaluated

    Returns
    -------
    bool
        Whether all items have the same length

    See Also
    --------
    check_len_equal :
        Identical functionality, but additionally returns the length of the
        input arguments
    """
    length1 = len(item1)

    for item in [item2] + list(args):
        if not len(item) == length1:
            return False

    return True


def max_list_item_len(input_list: Iterable) -> int:
    """Finds the maximum length of any item in a list or tuple

    Parameters
    ----------
    input_list : list or tuple
        Array of items to process

    Returns
    -------
    int
        Length of item in list with maximum length
    """
    return max(list(map(len, input_list)))



"""This module contains classes to perform unit conversions
"""

from .unit import (
    ConstantUnitMathConventions,
    Unit,
    UnitLinear,
    UnitLinearSI,
)
from .unitconverter import (
    UnitConverter,
    UnitConverterEntry,
    UnitConverterSI,
)
from .unitsystem import (
    UnitSystem,
    UnitSystemSI,
)



"""**Code for processing, parsing, and modifying strings**

The :py:mod:`pyxx.strings` module is intended to assist with processing data
in the form of Python strings.  Submodules provide functionality such as
analyzing strings with parentheses (e.g., determining which parentheses in
a string are "matched" with each other), assisting with string indexing
and splitting, and checking string content.
"""

from .functions import (
    contains_all_matched_brackets,
    find_matching_bracket,
    find_skip_brackets,
    split_at_index,
    str_excludes_chars,
    str_includes_only,
    strip_matched_brackets,
)



"""**Code for reading, writing, and processing files**

The :py:mod:`pyxx.files` module is intended to assist with processing
files.  It contains a set of classes intended to represent generic file
types (binary files and text files), as well as functions to perform
common file-related operations (such as computing file hashes).

An important concept about the classes in :py:mod:`pyxx.files` is that
files are considered *a data structure for storing information as a sequence
of characters*.  Files are not necessarily linked to a location on the disk
(although they can be).  Therefore, classes in :py:mod:`pyxx.files` can, but
do not necessarily, have an assigned path.

The intention of this module is that by subclassing the included classes,
the basic structure should be provided to manage file content, and a few very
general functions are provided.  Developers can add custom methods to
subclasses *specific to their file format*.  For instance, if creating a
``ShellScript`` subclass to parse shell scripts, a ``ShellScript.read()``
method might be added that parses the file, extracting the shebang line and
storing the interpreter in an attribute ``ShellScript.interpreter``.

Thus, the focus of the :py:mod:`pyxx.files` classes and subclasses should,
in most cases, be on processing context-specific file content, and less
focused on specific files on the disk (Python's built-in ``pathlib``,
``shutil``, and ``os`` modules are already well-suited to these purposes).
"""

from .classes import (
    BinaryFile,
    File,
    TextFile,
)
from .functions import compute_file_hash



"""The :py:class:`TypedList` class and inherited classes are intended to
provide nearly the same functionality of a Python :py:class:`list`, with the
added requirement that all items in the list are of a specific, user-specified
type.
"""

import itertools
import sys
from typing import Any, overload, Type, TypeVar

from pyxx.arrays.functions.size import max_list_item_len

if sys.version_info < (3, 9):
    from typing import Iterable, MutableSequence  # pragma: no cover
else:
    from collections.abc import Iterable, MutableSequence  # pragma: no cover


T = TypeVar('T')


class TypedList(MutableSequence[T]):
    """A list whose elements must be of a specific type

    This class defines a custom version of Python's :py:class:`list` object.
    When defining the :py:class:`TypedList` instance, users are required to
    specify a type, and all items added to the list must be of this type.
    Most of the attributes and methods of the built-in :py:class:`list` object
    can be used to manage data in a :py:class:`TypedList` instance.

    Examples
    --------
    Create an empty list of integers and add items to it using typical
    :py:class:`int` methods:

    >>> list_of_int = pyxx.arrays.TypedList(list_type=int)
    >>> list_of_int.append(0)
    >>> list_of_int.append(2)
    >>> list_of_int.extend([4, 6, 8])
    >>> print(list_of_int)
    [0, 2, 4, 6, 8]

    However, if we try to add items that aren't of type :py:class:`int`, an
    exception is thrown:

    >>> list_of_int.append(1.23)
    Traceback (most recent call last):
    ...
    TypeError: Item 1.23 is not of type <class 'int'>

    It's also possible to convert an existing list into a :py:class:`TypedList`:

    >>> current_list = [1, 2, 3, 4, 5]
    >>> typed_list = pyxx.arrays.TypedList(*current_list, list_type=int)
    >>> print(typed_list)
    [1, 2, 3, 4, 5]
    """

    def __init__(self, *values: T, list_type: Type[T],
                 print_multiline: bool = False,
                 multiline_padding: int = 1) -> None:
        """Creates a :py:class:`TypedList` list instance

        Creates an instance of the :py:class:`TypedList` object.  This list
        functions nearly identically to Python's built-in :py:class:`list`, but
        it requires that all items in the list be of a user-specified type.

        Parameters
        ----------
        *values : T
            Items to add to the list when initializing the list
        list_type : Type[T]
            The required type of all items in the list
        print_multiline : bool, optional
            Whether to return a printable string representation of the list in
            multiline format (default is ``False``).  Multiline format places
            each item in the list on its own line
        multiline_padding : int, optional
            The amount of horizontal padding to place between brackets and
            list items when creating a printable string representation in
            multiline format (default is ``1``).  Only applicable if
            ``print_multiline`` is ``True``
        """
        super().__init__()

        # Define internal list to store elements
        self._data: list = []

        # Store required type of all list elements
        self._list_type = list_type

        # Set default options when displaying object representation
        self._print_multiline = print_multiline
        self._multiline_padding = multiline_padding

        # Add any provided items to the list
        for item in values:
            self.append(item)

    def __check_item_type(self, value: Any) -> None:
        """Checks whether an item is of the type required for all items
        in the list, throwing an error if it is not"""
        if not isinstance(value, self.list_type):
            raise TypeError(f'Item {value} is not of type {self.list_type}')

    def __delitem__(self, index) -> None:
        """Deletes an item at a specified index from the list"""
        del self._data[index]

    def __eq__(self, value: object) -> bool:
        """Checks whether two :py:class:`TypedList` objects are equal (same
        list item type, same length, and all elements are equal)"""
        if not isinstance(value, TypedList):
            return False

        if self.list_type is not value.list_type:
            return False

        if not len(self) == len(value):
            return False

        for i, item in enumerate(self):
            if not item == value[i]:
                return False

        return True

    @overload
    def __getitem__(self, index: int) -> T:
        ...  # pragma: no cover

    @overload
    def __getitem__(self, index: slice) -> MutableSequence[T]:
        ...  # pragma: no cover

    def __getitem__(self, index):
        """Retrieves an item at a specified index from the list"""
        return self._data[index]

    def __len__(self) -> int:
        """Returns the number of items in the list"""
        return len(self._data)

    def __repr__(self) -> str:
        """Create a string representation illustrating the list and its
        contents"""
        if self.print_multiline:
            padding = ' ' * self.multiline_padding

            output = '['
            for i, item in enumerate(self._data):
                # If string representation of `item` has multiple lines,
                # add padding to each line
                item_formatted = str(item).replace('\n', '\n ' + padding)

                # Append each item in the list to the output string
                output += ((('\n ' if i > 0 else '')
                            + f'{padding}{str(item_formatted)},'))

            end_padding = max_list_item_len(output.split('\n')) \
                - len(output.split('\n')[-1]) + self.multiline_padding

            return output + f'{" " * end_padding}]'

        return str(self._data)

    @overload
    def __setitem__(self, index: int, value: T) -> None:
        ...  # pragma: no cover

    @overload
    def __setitem__(self, index: slice, value: Iterable[T]) -> None:
        ...  # pragma: no cover

    def __setitem__(self, index, value):
        """Sets the value of items at one or more indices in the list

        Like the built-in :py:class:`list` object, this method can either set
        a single item at a single specified index, or replace a slice given by
        ``index`` with the items in an iterable specified by ``value``.
        """
        # Check input types
        if isinstance(index, int):
            self.__check_item_type(value)
        elif isinstance(index, slice) \
                and isinstance(value, Iterable):
            for val in value:
                self.__check_item_type(val)
        else:
            raise TypeError(
                'Incompatible index-value type combination. Either provide a '
                'single index and a single value, or a slice of indices '
                'and a corresponding iterable of values')

        # Store items in list
        self._data[index] = value

    def __str__(self) -> str:
        """Returns a string representation of the list"""
        return self.__repr__()

    def insert(self, index: int, value: T):
        """Inserts a value at a given index in the list"""
        self.__check_item_type(value)
        self._data.insert(index, value)

    @property
    def list_type(self) -> Type[T]:
        """The required type of all items in the list"""
        return self._list_type

    @property
    def print_multiline(self) -> bool:
        """Whether to display the object's string representation in
        multiline format"""
        return self._print_multiline

    @print_multiline.setter
    def print_multiline(self, print_multiline: bool):
        if not isinstance(print_multiline, bool):
            raise TypeError(
                'Argument "print_multiline" must be of type "bool"')

        self._print_multiline = print_multiline

    @property
    def multiline_padding(self) -> int:
        """The number of spaces on either side of the list of items in
        the list when displaying the list's string representation in
        multiline format"""
        return self._multiline_padding

    @multiline_padding.setter
    def multiline_padding(self, multiline_padding: int):
        if type(multiline_padding) is not int:  # pylint: disable=C0123
            raise TypeError(
                'Argument "multiline_padding" must be of type "int"')
        self._multiline_padding = multiline_padding


class TypedListWithID(TypedList[T]):
    """A list whose elements must be of a specific type, where each class
    instance is given a unique identifier

    This class modifies the :py:class:`TypedList` class, assigning a unique
    identification number to each class instance.  This can be useful in
    tasks such as creating unique class instance names.

    Notes
    -----
    The identification number is shared by all instances of the
    :py:class:`TypedListWithID` class *and inherited classes*.  To reset
    the identifier for an inherited class, simply override the ``_id``
    iterator by defining ``_id = itertools.count(0)`` as a class variable
    for the subclass.
    """

    # Iterator that assigns a unique identification number to each
    # class instance
    _id = itertools.count(0)

    def __init__(self, *values: T, list_type: Type[T],
                 print_multiline: bool = True,
                 multiline_padding: int = 1) -> None:
        super().__init__(
            *values,
            list_type=list_type,
            print_multiline=print_multiline,
            multiline_padding=multiline_padding
        )

        # Set class instance identifier
        self.__id = next(self._id)

    @property
    def id(self) -> int:
        """A unique class instance identification number"""
        return self.__id



"""Code to assist in analyzing string content
"""


def str_excludes_chars(value: str, prohibited_chars: str) -> bool:
    """Checks that a string does not contain any of a specified list of
    prohibited characters

    Parameters
    ----------
    value : str
        String whose contents are to be analyzed
    prohibited_chars : str
        Characters which, if found in ``value``, should result in
        the function returning ``False``

    Returns
    -------
    bool
        Returns ``True`` if none of the characters in ``prohibited_chars`` are
        found in ``value``, and ``False`` otherwise

    Examples
    --------
    >>> pyxx.strings.str_excludes_chars('abcdefg', 'xyz')
    True

    >>> pyxx.strings.str_excludes_chars('abcdefg', 'a')
    False
    """
    if not (isinstance(value, str) and isinstance(prohibited_chars, str)):
        raise TypeError('Arguments must be of type "str"')

    return len(set(value).intersection(set(prohibited_chars))) <= 0


def str_includes_only(value: str, allowed_chars: str) -> bool:
    """Checks that a string contains only characters present in a specified
    set of characters

    Parameters
    ----------
    value : str
        String whose contents are to be analyzed
    allowed_chars : str
        If ``value`` contains characters that are not in ``allowed_chars``,
        the function should return ``False``

    Returns
    -------
    bool
        Returns ``True`` if ``value`` is composed of only characters found
        in ``allowed_chars``, and ``False`` otherwise
    """
    if not (isinstance(value, str) and isinstance(allowed_chars, str)):
        raise TypeError('Arguments must be of type "str"')

    return set(value).issubset(set(allowed_chars))



"""This module provides a data structure in which a collection of units
belonging to a common system of units can be stored, each with a unique
identifier.  This structure also allows metadata, such as tags and a
description, to be stored.
"""

import copy
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from pyxx.arrays.classes.typedlist import TypedList
from pyxx.arrays.functions.convert import convert_to_tuple
from pyxx.units.exceptions import (
    IncompatibleUnitsError,
    InvalidSearchFieldError,
    UnitAlreadyDefinedError,
    UnitNotFoundError,
)
from pyxx.arrays.functions.size import (
    check_len_equal,
    max_list_item_len,
)
from pyxx.units.functions.parser import parse_unit
from .constants import SAMPLE_SI_UNITS
from .unit import Unit
from .unitsystem import UnitSystem, UnitSystemSI


class UnitConverterEntry:
    """Data structure for storing information about a unit in a unit
    converter

    The :py:class:`UnitConverterEntry` is designed to be used in combination
    with a :py:class:`UnitConverter` object.  A :py:class:`UnitConverterEntry`
    object stores data about a particular unit, allowing such information to
    be grouped and changes to the data validated.  Many such objects are
    stored in in a :py:class:`UnitConverter`, allowing it to relate and perform
    conversions between different units, while the "lower-level" details of
    managing unit data are handled by :py:class:`UnitConverterEntry`.
    """

    def __init__(self, unit: Unit,
                 tags: Optional[
                     Union[TypedList[str], List[str], Tuple[str, ...], str]
                 ] = None,
                 name: Optional[str] = None,
                 description: Optional[str] = None) -> None:
        """Creates a :py:class:`UnitConverterEntry` to store data about a
        particular unit

        Parameters
        ----------
        unit : Unit
            A :py:class:`Unit` object describing the unit
        tags : TypedList or list or tuple or str, optional
            One or more strings containing brief, one-word descriptors to use
            to group similar units, such as "length" or "speed" (default is
            ``None``, meaning that no tags are associated with the unit)
        name : str, optional
            A name for the unit (default is ``None``)
        description : str, optional
            A written description of the unit (default is ``None``)
        """
        self.description = description
        self.name = name
        self.tags = tags  # type: ignore
                          # (workaround for python/mypy#3004)  # noqa: E114, E116
        self.unit = unit

    def __repr__(self) -> str:
        representation = f'{self.__class__}'

        if self.name is not None:
            representation += f'\n-- Name: {self.name}'

        if self.description is not None:
            representation += f'\n-- Description: {self.description}'

        if len(self.tags) > 0:
            tags = copy.deepcopy(self.tags)
            tags.print_multiline = False
            representation += f'\n-- Tags: {tags}'

        representation += f'\n-- Unit: {self.unit}'

        return representation

    def __str__(self) -> str:
        return self.__repr__()

    @property
    def description(self) -> Union[str, None]:
        """A description of the unit"""
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        if description is None:
            self._description = None
        else:
            self._description = str(description)

    @property
    def name(self) -> Union[str, None]:
        """A name by which to reference the unit"""
        return self._name

    @name.setter
    def name(self, name: Optional[str]) -> None:
        if name is None:
            self._name = None
        else:
            self._name = str(name)

    @property
    def tags(self) -> TypedList[str]:
        """Tags that associate the unit with various categories or groups"""
        return self._tags

    @tags.setter
    def tags(self,
             tags: Optional[
                 Union[TypedList[str], List[str], Tuple[str, ...], str]
             ]) -> None:

        if isinstance(tags, TypedList) and (tags.list_type is str):
            self._tags = tags

        elif tags is None:
            self._tags = TypedList(list_type=str)

        elif isinstance(tags, str):
            self._tags = TypedList(tags, list_type=str)

        elif isinstance(tags, (list, tuple)):
            try:
                self._tags = TypedList(*tags, list_type=str)
            except (TypeError, ValueError) as exception:
                raise TypeError('Unable to set unit tags.  All tags must be of '
                                'type "str"') from exception

        else:
            raise TypeError('Argument "tags" is not a valid type')

        self._tags.print_multiline = False

    @property
    def unit(self) -> Unit:
        """A :py:class:`Unit` object representing the unit"""
        return self._unit

    @unit.setter
    def unit(self, unit: Unit) -> None:
        if not isinstance(unit, Unit):
            raise TypeError(f'Argument "unit" must be of type {Unit}')

        self._unit = unit


class UnitConverter(Dict[str, UnitConverterEntry]):
    """Data structure for storing units and performing unit conversions

    This class is intended to store a collection of units, and provide methods
    to convert quantities between the stored units or any valid arithmetic
    combination of stored units.  There are no limit on the number or types of
    units that can be stored; however, all units must belong to the same
    system of units.

    The :py:class:`UnitConverter` class inherits from Python's built-in
    :py:class:`dict` object, so any of the methods that are defined for a
    :py:class:`dict` should also work for :py:class:`UnitConverter` objects.

    Examples
    --------
    For examples, refer to the :ref:`section-tutorials_units` page.
    """

    def __init__(self, unit_system: UnitSystem):
        """Creates a new unit converter

        Creates an new :py:class:`UnitConverter` class with no units defined.

        Parameters
        ----------
        unit_system : UnitSystem
            The system of units to which all units in the unit converter must
            belong
        """
        # Validate and store inputs
        if not isinstance(unit_system, UnitSystem):
            raise TypeError('Argument "unit_system" must be an instance or '
                            f'subclass of {UnitSystem}')
        self._unit_system = unit_system

    def __getitem__(self, key: str) -> UnitConverterEntry:
        if not isinstance(key, str):
            raise TypeError('Argument "key" must be of type "str"')

        if key not in self:
            raise UnitNotFoundError(
                f'Unit "{key}" was not found in unit converter')

        return super().__getitem__(key)

    def __repr__(self) -> str:
        return (f'{self.__class__}\n'
                f'-- System of units: {self.unit_system}\n'
                f'{str(self)}')

    def __setitem__(self, key: str, value: UnitConverterEntry):
        if not isinstance(value, UnitConverterEntry):
            raise TypeError(
                f'Unit converter units must be of type {UnitConverterEntry}. '
                'Consider using the `.add_unit()` method to add units')

        # Validate `key`
        if not isinstance(key, str):
            raise TypeError(
                'Unit converter unit identifier keys must be of type "str"')

        if not self.is_simplified_unit(key):
            raise ValueError(
                f'Unit identifier key "{key}" is not a fully-simplified unit')

        # Verify that system of units of `value` matches unit converter
        #   This does not use `isinstance()` or else units with system of units
        #   `UnitSystem` could be converted to any other system of units
        if type(value.unit.unit_system) is not type(self.unit_system):  # noqa: E721
            raise TypeError(
                'System of units used by unit converter and new unit '
                f'differ: unit converter uses {type(self.unit_system)} '
                f'while new unit uses {type(value.unit.unit_system)}')

        super().__setitem__(key, value)

    def __str__(self) -> str:
        return '\n'.join(self._generate_unit_table(list(self.keys())))

    def _generate_unit_table(self, keys: Union[List[str], Tuple[str, ...]],
                             col_spacing: int = 4,
                             generate_header: bool = True,
                             trim_trailing_whitespace: bool = True
                             ) -> List[str]:
        # Generate lists of data to return
        def copy_tags_disable_multiline(tags: TypedList[str]):
            tags_copy = copy.deepcopy(tags)
            tags_copy.print_multiline = False
            return tags_copy

        unit_table_col_data = {
            'Key': [str(key) for key in keys],
            'Name': [str(self[key].name) for key in keys],
            'Tags': [str(copy_tags_disable_multiline(self[key].tags))
                     for key in keys],
            'base_unit_exps': [str(self[key].unit.base_unit_exps)
                               for key in keys],
            'Description': [str(self[key].description) for key in keys],
        }

        # Verify that all lists are of the same length
        is_equal, num_keys = check_len_equal(*unit_table_col_data.values())

        if not is_equal:
            # This error should never be raised (if it is, there was a
            # significant issue with the structure of the data in the unit
            # converter), but it is included as a backup check
            raise ValueError(  # pragma: no cover
                'Unit converter data appears to be corrupted. Please remove '
                'and replace all units, or set up a new unit converter')

        # Determine width of each column to return
        unit_table_col_widths = {}
        for key, value in unit_table_col_data.items():
            if len(value) == 0:
                unit_table_col_widths[key] = len(key)
            else:
                unit_table_col_widths[key] \
                    = max(max_list_item_len(value), len(key))

        # Generate table header
        unit_table: List[str] = []

        if generate_header:
            # Column names
            line = ''
            for i, column_name in enumerate(unit_table_col_data.keys()):
                # Determine formatting settings
                num_trailing_spaces = col_spacing \
                    if i < len(unit_table_col_data) - 1 else 0
                print_width = \
                    unit_table_col_widths[column_name] + num_trailing_spaces

                # Construct next line in table
                line += f'{column_name:{print_width}s}'

            unit_table.append(line)

            # Header underline
            underline_len = sum(unit_table_col_widths.values()) \
                + (col_spacing * (len(unit_table_col_data) - 1))
            unit_table.append('-' * underline_len)

        # Add units to table
        for j in range(num_keys):
            line = ''
            for i, (column_name, column_content) \
                    in enumerate(unit_table_col_data.items()):
                # Determine formatting settings
                num_trailing_spaces = col_spacing \
                    if i < len(unit_table_col_data) - 1 else 0
                print_width = \
                    unit_table_col_widths[column_name] + num_trailing_spaces

                # Construct next line in table
                line += f'{column_content[j]:{print_width}s}'

            unit_table.append(line)

        # If requested, remove trailing whitespace from each line
        if trim_trailing_whitespace:
            unit_table = [line.rstrip() for line in unit_table]

        return unit_table

    @property
    def unit_system(self) -> UnitSystem:
        """The system of units used by all units in the unit converter"""
        return self._unit_system

    def add_unit(self, key: str, unit: Unit,
                 tags: Optional[
                     Union[TypedList[str], List[str], Tuple[str, ...], str]
                 ] = None,
                 name: Optional[str] = None, description: Optional[str] = None,
                 aliases:
                     Optional[Union[str, List[str], Tuple[str, ...]]] = None,
                 overwrite: bool = False) -> None:
        """Adds a unit to the unit converter

        Adds a unit and possibly unit metadata (tags, description) to the unit
        converter, so that the unit can be used to perform later unit
        conversions.

        Parameters
        ----------
        key : str
            A short string, such as ``'kg'``, uniquely identifying the unit
        unit : Unit
            A :py:class:`Unit` object (or subclass) specifying the unit
        tags : TypedList or list or tuple or str, optional
            One or more strings containing brief, one-word descriptors to use
            to group similar units, such as "length" or "speed" (default is
            ``None``, meaning that no tags are associated with the unit)
        name : str, optional
            A name for the unit (default is ``None``)
        description : str, optional
            A written description of the unit (default is ``None``)
        aliases : str or list or tuple, optional
            Alternate identifiers to add to unit converter (default is
            ``None``).  Refer to the :py:meth:`add_alias` method documentation
            for more detail about unit aliases
        overwrite : bool, optional
            Whether to overwrite units if they already exist in the unit
            converter (default is ``False``)
        """
        if key in self and not overwrite:
            raise UnitAlreadyDefinedError(
                f'Key "{key}" already exists. To automatically overwrite, '
                'call `.add_unit()` with `overwrite=True`')

        self[key] = UnitConverterEntry(unit=unit, tags=tags, name=name,
                                       description=description)

        if aliases is not None:
            self.add_alias(key, aliases)

    def add_alias(self, key: str,
                  aliases: Union[str, List[str], Tuple[str, ...]],
                  overwrite: bool = False) -> None:
        """Add one or more aliases for a unit

        Adds alternate identifiers (keys) that may be used to identify a unit in
        simple or compound unit strings.  This method is useful for adding
        units which have multiple equivalent representations, such as "micron"
        and ":spelling:word:`μm`," to the unit converter efficiently.

        Parameters
        ----------
        key : str
            The unit identifier of a unit already in the unit converter for
            which to add an alias
        aliases : str or list or tuple
            Alternate identifiers to add to unit converter
        overwrite : bool, optional
            Whether to overwrite units if they already exist in the unit
            converter (default is ``False``)

        Warnings
        --------
        Aliased units are added to the :py:class:`UnitConverter` **by
        reference**.  Thus, if one of the aliased units is modified, all the
        other aliased units will also be updated.
        """
        for alias in convert_to_tuple(aliases):
            if alias in self and not overwrite:
                raise UnitAlreadyDefinedError(
                    f'Key "{alias}" already exists. To automatically '
                    'overwrite, set `overwrite=True`')

            self[alias] = self[key]

    def convert(self, quantity: Union[np.ndarray, list, tuple, float],
                from_unit: str, to_unit: str) -> np.ndarray:
        """Converts a quantity from one unit in the :py:class:`UnitConverter`
        to another

        Converts a value or array from one unit to another, where the units
        can be either units in the unit converter, or compound units composed
        of units in the unit converter.

        Parameters
        ----------
        quantity : np.ndarray or list or tuple or float
            Value(s) to convert from ``from_unit`` to ``to_unit``
        from_unit : str
            String specifying the simple or compound unit of ``quantity``
        to_unit : str
            String specifying the simple or compound unit to which
            ``quantity`` is to be converted

        Returns
        -------
        np.ndarray
            NumPy array with the same shape as ``quantity`` containing the
            value(s) in ``quantity`` converted from ``from_unit`` to
            ``to_unit``
        """
        if not (isinstance(from_unit, str) and isinstance(to_unit, str)):
            raise TypeError(
                'Arguments "from_unit" and "to_unit" must be strings '
                'corresponding to units defined in the unit converter')

        if not self.is_convertible(from_unit, to_unit):
            raise IncompatibleUnitsError(
                f'Unable to convert quantity: units "{from_unit}" and '
                f'"{to_unit}" are not compatible')

        # If input and output units are identical, directly return quantity
        if from_unit == to_unit:
            return np.array(quantity)

        # Convert to `Unit` objects
        from_unit_obj = self.str_to_unit(from_unit)
        to_unit_obj = self.str_to_unit(to_unit)

        return from_unit_obj.convert(quantity, 'to', to_unit_obj)

    def get_aliases(self, unit: str) -> List[str]:
        """Retrieves a list of aliases for a particular unit in the unit
        converter

        This method identifies all keys in the unit converter that are aliases
        of a particular unit.

        Parameters
        ----------
        unit : str
            The identifier of the unit for which to search for aliases

        Returns
        -------
        list
            A list of strings containing all identifiers in the unit converter
            which are aliases of ``unit``
        """
        if unit not in self:
            raise UnitNotFoundError(
                f'Unit "{unit}" is not defined in the unit converter')

        unit_id = id(self[unit])

        aliases = []
        for key in self:
            if (key != unit) and (id(self[key]) == unit_id):
                aliases.append(key)

        return aliases

    def is_convertible(self, unit1: str, unit2: str, *args: str) -> bool:
        """Checks whether units can be converted between each other

        Checks whether two or more units are compatible and can be directly
        converted between each other.  Inputs can be strings of simple or
        compound units.  The corresponding units (or component units of
        compound units) must exist in the :py:class:`UnitConverter` before
        calling this method.

        Parameters
        ----------
        unit1 : str
            Key/identifier of the first unit to check
        unit2 : str
            Key/identifier of the second unit to check
        *args : str, optional
            Keys/identifiers of any other units to check

        Returns
        -------
        bool
            Whether all units can be converted directly between each other
        """
        unit1_unit = self.str_to_unit(unit1)
        units_to_check = [self.str_to_unit(i) for i in ([unit2] + list(args))]

        for unit in units_to_check:
            if not unit1_unit.is_convertible(unit):
                return False

        return True

    def is_defined_unit(self, unit: str) -> bool:
        """Checks whether a simple or complex unit is composed exclusively of
        units which have been defined in the unit converter

        This method verifies that a simple unit or all component units of a
        compound unit have been defined in the unit converter; if not, unit
        conversions involving such a unit cannot be performed.

        Parameters
        ----------
        unit : str
            The simple or compound unit to analyze

        Returns
        -------
        bool
            Whether all component unit(s) in ``unit`` are defined in the
            unit converter
        """
        if unit.strip() == '':
            return False

        component_units = parse_unit(str(unit)).keys()

        for component in component_units:
            if component not in self:
                return False

        return True

    def is_simplified_unit(self, unit: str) -> bool:
        """Evaluates whether a unit is a fully-simplified unit

        Returns whether a unit is NOT a compound unit.  Fully-simplified units
        can be added to the :py:class:`UnitConverter`, but compound units
        cannot.  Note that units that are not of type :py:class:`str` or which
        are enclosed in brackets (such as ``'(m)'``) are not considered
        fully-simplified.

        Parameters
        ----------
        unit : str
            The unit to check

        Returns
        -------
        bool
            Whether ``unit`` is fully-simplified and can be added to the
            :py:class:`UnitConverter`
        """
        if not isinstance(unit, str):
            return False

        return parse_unit(unit) == {unit: 1.0}

    def list_tags(self) -> List[str]:
        """Returns a list of all tags currently used in the
        :py:class:`UnitConverter` instance

        Returns
        -------
        list
            List (in alphabetical order) of all tags used by any unit in the
            unit converter
        """
        tags: List[str] = []
        for entry in self.values():
            tags = tags + list(entry.tags)

        return sorted(list(set(tags)))

    def search(self, search_term: str,
               search_fields: Union[List[str], Tuple[str, ...], str]
                   = ('key', 'name', 'tags', 'description'),  # noqa: E127
               filter_by_tags:
                   Optional[Union[List[str], Tuple[str, ...], str]] = None,
               hide_aliases: bool = False,
               print_results: bool = True, return_results: bool = False
               ) -> Union[List[str], None]:
        """Searches the :py:class:`UnitConverter` contents for a given term

        Once a unit converter has a large number of units, it can be
        challenging to remember all included units.  This method provides a
        simple way to search the contents of a :py:class:`UnitConverter`,
        either interactively (on the command line) or programmatically (in a
        script).

        Parameters
        ----------
        search_term : str
            The term to search for.  Any units in which ``search_term`` is
            found in the fields specified by ``search_fields`` will be
            returned.  Set ``search_term`` to ``'*'`` or ``'**'`` to match
            any string
        search_fields : list or tuple or str, optional
            The fields of each entry in the :py:class:`UnitConverter` to search.
            Valid options are any of ``('key', 'name', 'tags', 'description')``
            (default is to search all valid fields)
        filter_by_tags : list or tuple or str, optional
            Only units with tags specified by ``filter_by_tags`` will be
            included in the search results.  Set to ``None`` disable filtering
            by tags (default is ``None``)
        hide_aliases : bool, optional
            Whether to hide unit aliases when outputting search results
            (default is ``False``).  Refer to the :py:meth:`add_alias` method
            documentation for more detail about unit aliases
        print_results : bool, optional
            Whether to print search results to the terminal (default is
            ``True``)
        return_results : bool, optional
            Whether to return a list containing search results (default is
            ``False``)

        Returns
        -------
        list or None
            Returns a list of strings containing the :py:class:`UnitConverter`
            keys for any units matching the specified search criteria if
            ``return_results`` is ``True``.  Otherwise, no outputs are
            returned
        """
        # Validate and pre-process inputs
        if not isinstance(search_term, str):
            raise TypeError('Argument "search_term" must be a string')

        if not isinstance(search_fields, (list, tuple)):
            search_fields = (search_fields,)

        allowed_search_fields = set(('key', 'name', 'tags', 'description'))
        if not set(search_fields).issubset(allowed_search_fields):
            raise InvalidSearchFieldError(
                'Invalid search fields: '
                f'{set(search_fields) - allowed_search_fields}')

        if not ((filter_by_tags is None)
                or isinstance(filter_by_tags, (list, tuple))):
            filter_by_tags = (filter_by_tags,)

        # Search for matching unit converter entries
        search_results = []
        search_results_ids = []
        for key, entry in self.items():
            # If aliases are to be hidden and the unit has already been saved
            # to the search results, skip unit
            if hide_aliases and (id(entry) in search_results_ids):
                continue

            # If filtering by tags and none of the tags corresponding to "key"
            # are to be shown, skip unit
            if filter_by_tags is not None:
                filter_tags = set(list(filter_by_tags))
                unit_tags = set(list(entry.tags))

                if len(filter_tags.intersection(unit_tags)) <= 0:
                    continue

            if search_term in ('*', '**'):
                search_results.append(key)
                search_results_ids.append(id(entry))
                continue

            if 'key' in search_fields:
                if search_term in key:
                    search_results.append(key)
                    search_results_ids.append(id(entry))
                    continue

            if 'name' in search_fields:
                if (entry.name is not None) and (search_term in entry.name):
                    search_results.append(key)
                    search_results_ids.append(id(entry))
                    continue

            if 'tags' in search_fields:
                if len(entry.tags) > 0:
                    results = []
                    for tag in entry.tags:
                        results.append(search_term in tag)

                    if any(results):
                        search_results.append(key)
                        search_results_ids.append(id(entry))
                        continue

            # This "if" statement should never be false (it was already checked
            # that the contents of "search_fields" were from an expected list),
            # but this redundant check is retained as a backup
            if 'description' in search_fields:  # pragma: no cover
                if (entry.description is not None) \
                        and (search_term in entry.description):
                    search_results.append(key)
                    continue

        # Output search results
        if print_results:
            print('\n'.join(self._generate_unit_table(search_results)))

        if return_results:
            return search_results

        return None

    def str_to_unit(self, unit: str) -> Unit:
        """Converts a string to a :py:class:`Unit` object using units
        defined in the :py:class:`UnitConverter`

        Parses a string containing a simple or compound unit, converting it
        to a :py:class:`Unit` object.  All simple units and all component
        units in compound units must be defined in the unit converter prior
        to calling this method.

        Parameters
        ----------
        unit : str
            String containing the simple or compound unit to parse

        Returns
        -------
        Unit
            A :py:class:`Unit` object representation of the unit specified
            by ``unit``
        """
        if unit.strip() == '':
            raise UnitNotFoundError(
                f'Unit "{unit}" consists only of whitespace and is not valid')

        parsed_unit = parse_unit(unit)

        # For special case of dimensionless units, output a constant unit
        if len(parsed_unit) == 0:
            return Unit(
                unit_system        = self.unit_system,
                base_unit_exps     = [0] * self.unit_system.num_base_units,
                to_base_function   = lambda x, exp: x,
                from_base_function = lambda x, exp: x
            )

        # Initialize output unit using one of the units in `parsed_unit`
        unit_str0, exp0 = parsed_unit.popitem()
        output_unit = (self[unit_str0].unit)**exp0

        # Add all remaining units in `parsed_unit`
        for unit_str, exp in parsed_unit.items():
            output_unit *= (self[unit_str].unit)**exp

        return output_unit


class UnitConverterSI(UnitConverter):
    """A unit converter using the SI system of units, pre-filled with
    common units

    This class is identical in function to a :py:class:`UnitConveter`, with
    the exception that it has been initialized with a number of commonly-used
    units.  All units in a :py:class:`UnitConveterSI` are defined in terms of
    the SI system (https://www.nist.gov/pml/owm/metric-si/si-units), with the
    following sequence of seven base units (used in the
    :py:attr:`Unit.base_unit_exps` attribute of all units in the unit
    converter):

    1. Length: meter [:math:`m`]
    2. Time: second [:math:`s`]
    3. Amount of substance: mole [:math:`mol`]
    4. Electric current: ampere [:math:`A`]
    5. Temperature: Kelvin [:math:`K`]
    6. Luminous intensity: candela [:math:`cd`]
    7. Mass: kilogram [:math:`kg`]

    For a list of units available in this unit converter, refer to the
    :ref:`section-unitconverter_units` page.
    """

    def __init__(self):
        """Creates a new unit converter using the SI system of units and
        populated with commonly-used units.

        Creates an new :py:class:`UnitConverterSI` instance.  This unit
        converter uses the SI system (specifically, the system of units
        referenced by :py:class:`UnitSystemSI`) and is pre-filled with a
        set of commonly-used units, listed on the
        :ref:`section-unitconverter_units` page.
        """
        super().__init__(unit_system=UnitSystemSI())

        for key, unit_data in SAMPLE_SI_UNITS.items():
            self.add_unit(key=key, **unit_data)



"""**Objects for performing unit conversions**

The :py:mod:`pyxx.units` module provides unit conversion capabilities.  It
allows users to define customized systems of units and add associated,
arbitrary units and convert quantities between such units.
"""

from .classes import (
    ConstantUnitMathConventions,
    Unit,
    UnitConverter,
    UnitConverterEntry,
    UnitConverterSI,
    UnitLinear,
    UnitLinearSI,
    UnitSystem,
    UnitSystemSI,
)
from .classes.cli import execute_from_command_line
from .functions import parse_unit



"""Customized exceptions for the :py:mod:`pyxx.strings` module
"""


class BracketErrors(Exception):
    """Errors for the ``pyxx.strings.brackets`` module"""

class NotABracketError(BracketErrors, ValueError):
    """Error thrown if encountering a non-bracket character when
    expecting to encounter a bracket character"""

class UnmatchedBracketsError(BracketErrors, ValueError):
    """Error thrown if unmatched brackets are found in an
    unexpected context"""



"""This module is used to process text files
"""

import copy
import pathlib
from typing import List, Optional, Tuple, Union

from pyxx.arrays.functions.convert import convert_to_tuple
from .file import File


class TextFile(File):
    """Base class for processing text files

    This class can be used to represent text files (that is, files with
    a series of ASCII-based characters as content, that can be open and
    read with an editor such as Notepad).  It provides the capability to
    read/write text files and perform processing operations such as
    removing commented lines.
    """

    def __init__(self, path: Optional[Union[str, pathlib.Path]] = None,
                 comment_chars: Optional[Union[Tuple[str, ...], str]] = None
                 ) -> None:
        """Define a text file

        Creates an object that represents and can be used to process
        a text file.

        Parameters
        ----------
        path : str or pathlib.Path, optional
            Location of the text file in the file system  (default is ``None``)
        comment_chars : tuple or str, optional
            Character(s) considered to represent comments in the text file
            (default is ``None``, which considers no characters to denote
            comments in the file)

        Notes
        -----
        Passing an empty string (``''``) or empty tuple (``()``) as the
        ``comment_chars`` argument is equivalent to passing ``None`` (or
        not providing this argument) -- in all these cases, the file will
        be considered to have no characters denoting comments.
        """
        super().__init__(path=path)

        # Mypy exclusions in constructor were added as workarounds
        # for python/mypy#3004

        # Initialize lists to store file content
        self._contents: List[str] = []
        self._raw_contents: Union[List[str], None] = None

        # Initialize file properties
        self._line_ending: Union[str, Tuple[str, ...], None] = None
        self.trailing_newline = None  # type: ignore

        # Store comment character
        self.comment_chars = comment_chars  # type: ignore

    def _check_contents(self, contents: List[str]) -> None:
        # Verify that input is a list
        if not isinstance(contents, list):
            raise TypeError('Argument "contents" must be of type "list"')

        # Verify that all elements of the input are strings
        for line in contents:
            if not isinstance(line, str):
                raise TypeError(
                    'All elements of "contents" must be of type "str"')

    @property
    def comment_chars(self) -> Union[Tuple[str, ...], None]:
        """A tuple of all characters considered to denote comments"""
        return self._comment_chars

    @comment_chars.setter
    def comment_chars(self, comment_chars: Union[Tuple[str, ...], str, None]
                      ) -> None:
        if comment_chars is None:
            self._comment_chars = None
        else:
            if isinstance(comment_chars, (str, tuple)):
                if len(comment_chars) == 0:
                    self._comment_chars = None
                else:
                    comment_chars_list = convert_to_tuple(comment_chars)

                    # Make sure every character of comment characters tuple is
                    # a string
                    for item in comment_chars_list:
                        if not isinstance(item, str):
                            raise TypeError(
                                'All comment characters must be of type "str"')

                    self._comment_chars = comment_chars_list
            else:
                raise TypeError(
                    'Argument "comment_chars" must be either `None` or of '
                    'type "str" or "tuple"')

    @property
    def contents(self) -> List[str]:
        """A reference to a list containing the (potentially modified) file
        content of each line of the file

        Warnings
        --------
        This attribute returns the list **by reference**.  This means that
        if you set a variable equal to this reference, then editing this
        variable will edit the :py:attr:`contents` attribute (e.g., if you
        set ``my_content = MyTextFile.contents``, then editing ``my_content``
        will change the content stored in ``MyTextFile``).

        Notes
        -----
        If trying to set the :py:attr:`contents` attribute, do not try to set
        this attribute directly (i.e., don't use code similar to
        ``MyTextFile.contents = ['line1', 'line2', 'line3']``).  Instead, use
        the :py:meth:`set_contents` method, as it offers greater control
        over whether the contents are passed by reference or value.
        """
        return self._contents

    @property
    def line_ending(self) -> Union[str, Tuple[str, ...]]:
        """The character(s) used to denote the end of lines in the text file

        This property only applies to files that were read using the
        :py:meth:`read` method.  After reading a file, this property stores
        the line ending(s) used in the file.  Lines in text files can be
        terminated with ``'\\n'`` (LF), ``'\\r\\n'`` (CRLF), ``'\\r'``, or a
        combination of these characters (potentially with different line
        endings on different lines).

        After reading a file, this property stores either a string containing
        the line endings on every line of the file, or a tuple containing all
        line endings encountered throughout the file.
        """
        if self._line_ending is None:
            raise AttributeError(
                'Attribute "line_ending" has not been set. Please ensure '
                'that either the `read()` method has been called')

        return self._line_ending

    @property
    def raw_contents(self) -> Union[List[str], None]:
        """A copy of the raw file content

        If the file was read using the :py:meth:`read` method, this attribute
        stores the original, unaltered contents of each line of the input
        file, and it returns a copy of this list of lines.  If the file was
        not read with the :py:meth:`read` method, this attribute stores a
        value of ``None``.
        """
        return copy.deepcopy(self._raw_contents)

    @property
    def trailing_newline(self) -> bool:
        """Whether the original file had a newline at the end
        of the file"""
        if self._trailing_newline is None:
            raise AttributeError(
                'Attribute "trailing_newline" has not been set. Please ensure '
                'that either the `read()` or `set_contents()` method has '
                'been called')

        return self._trailing_newline

    @trailing_newline.setter
    def trailing_newline(self, trailing_newline: Union[bool, None]) -> None:
        if trailing_newline is None:
            self._trailing_newline = None
        else:
            self._trailing_newline = bool(trailing_newline)

    def clean_contents(self,
                       remove_comments: bool = False,
                       skip_full_line_comments: bool = False,
                       strip: bool = False,
                       concat_lines: bool = False,
                       remove_blank_lines: bool = False
                       ) -> None:
        """Clean :py:attr:`contents` in-place

        Cleans :py:attr:`contents` (removing comments, blank lines, etc.)
        based on user-defined rules.  Modifications are made in-place
        (i.e., the resulting content is stored in :py:attr:`contents`).

        Parameters
        ----------
        remove_comments : bool, optional
            Whether to remove comments from file (default is ``True``)
        skip_full_line_comments : bool, optional
            Whether to skip removing comments where the comment is the only
            text on a line.  Only applies if ``remove_comments`` is ``True``
            (default is ``False``)
        strip : bool, optional
            Whether to strip leading and trailing whitespace from each
            line (default is ``True``)
        concat_lines : bool, optional
            Whether to concatenate lines ending with a backslash with the
            following line (default is ``True``)
        remove_blank_lines : bool, optional
            Whether to remove lines that contain no content after other
            cleaning operations have completed (default is ``True``)
        """
        # Confirm that "contents" attribute hasn't been modified improperly
        self._check_contents(self.contents)

        # Store original file contents
        orig_contents = copy.deepcopy(self.contents)

        # Clean file line-by-line
        self._contents = []
        i = 0
        while (i < len(orig_contents)):
            line = orig_contents[i]

            # If line ends with "\", concatenate with next line
            if concat_lines:
                while (line.strip().endswith('\\') and (i < len(orig_contents))):
                    line = line.rsplit('\\', maxsplit=1)[0] + orig_contents[i+1]
                    i += 1

            # Remove comments
            if remove_comments and (self.comment_chars is not None):
                if not (skip_full_line_comments
                        and line.strip().startswith(self.comment_chars)):
                    for comment_char in self.comment_chars:
                        line = line.split(comment_char, maxsplit=1)[0]

            # Strip whitespace from beginning and end of line
            if strip:
                line = line.strip()

            # Remove blank lines
            if not (remove_blank_lines and len(line.strip()) == 0):
                self._contents.append(line)

            i += 1

    def overwrite(self, prologue: str = '', epilogue: Optional[str] = None,
                  line_ending: str = '\n') -> None:
        """Write data in :py:attr:`contents` to the file specified by
        :py:attr:`path`

        Writes the lines of content in the :py:attr:`contents` attribute to
        the (previously-defined) file specified by the :py:attr:`path`
        attribute, suppressing warnings before overwriting the file.  This
        is useful for cases when the file contents are manually populated and
        it is desired to "dump" them to a file.  This method is also useful if
        a file's contents need to be updated periodically based on the results
        of another process.

        Parameters
        ----------
        prologue : str, optional
            Content written at beginning of file (default is ``''``)
        epilogue : str, optional
            Content written at end of file (default is to use the value of the
            ``line_ending`` argument if :py:attr:`trailing_newline` is
            ``True`` and ``''`` otherwise)
        line_ending : str, optional
            String written at the end of each line when writing file content
            (default is ``'\\n'``)
        """
        if self.path is None:
            raise AttributeError('Attribute "path" must be set to call '
                                 '"overwrite()" method')

        self.write(
            output_file = self.path,
            write_mode = 'w',
            warn_before_overwrite = False,
            prologue = prologue,
            epilogue = epilogue,
            line_ending = line_ending
        )

    def parse(self) -> None:
        """Parses the data in :py:attr:`contents` and stores it in class
        attributes

        This method by default does nothing.  However, it is intended that
        subclasses of :py:class:`TextFile` should override this method and
        define file-specific behavior in this method for extracting data from
        the file and storing it in custom object attributes.

        For example, if defining a CSV-parser, the :py:meth:`parse` method
        might parse data from the file and store it as a NumPy array.
        """
        return None

    def read(self, path: Optional[Union[str, pathlib.Path]] = None,
             parse: bool = True) -> None:
        """Read file from disk

        Calling this method reads the file specified by the :py:attr:`path`
        attribute from the disk, populating :py:attr:`contents` and
        :py:attr:`raw_contents`.  Additionally, the file hashes stored in
        the :py:attr:`hashes` attribute are updated (to make it easier to
        check if the file has been modified later).

        Parameters
        ----------
        path : str or pathlib.Path, optional
            Location of the text file in the file system  (default is ``None``)
        parse : bool, optional
            Whether to call the :py:meth:`parse` method after reading the
            file (default is ``True``)
        """
        # Mypy type annotation added because immediately after calling
        # `set_read_metadata()`, the "path" attribute cannot be `None`
        # or else an error would have been thrown
        self.set_read_metadata(path)
        self.path: pathlib.Path

        # Read file
        with open(self.path, 'r', encoding='utf_8') as fileID:
            self._raw_contents = fileID.readlines()

            # Store line endings
            self._line_ending = fileID.newlines

        # Store whether original file has a trailing newline
        self.trailing_newline = self._raw_contents[-1].endswith('\n')

        # Remove trailing newlines.  This is beneficial because if the
        # file is later cleaned and, for example, comments are removed,
        # this can result in an unpredictable mix of lines with trailing
        # newlines and lines without, so it's simpler to remove them all
        # at the beginning and add them when writing the file
        self._contents = [line.rstrip('\r\n') for line in self._raw_contents]

        # Optionally parse file contents
        if parse:
            self.parse()

    def set_contents(self, contents: List[str], trailing_newline: bool,
                     pass_by_reference: bool = False) -> None:
        """Add data to the :py:attr:`contents` list

        Allows users to manually fill the :py:attr:`contents` list with
        user-defined content.  The input list must be a list of strings,
        and the user can optionally choose whether to pass the input by
        reference or value.

        Parameters
        ----------
        contents : list
            List of strings which are to be assigned to the
            :py:attr:`contents` list
        trailing_newline : bool
            Whether the contents being added represent a file with a trailing
            newline (because the file wasn't read, the object has no way to
            determine whether the file has a trailing newline, so users must
            provide this information)
        pass_by_reference : bool, optional
            Whether to pass the ``contents`` argument by reference (default
            is ``False``)

        Notes
        -----
        If passing ``contents`` by reference, this means that if subsequent
        changes are made to the original ``contents`` object, they will be
        reflected in the :py:attr:`contents` attribute.  If passing by value,
        then a *copy* of the ``contents`` argument will be made, so changing
        the object outside the class instance will not affect the
        :py:attr:`contents` attribute.
        """
        # Verify that input matches required format
        self._check_contents(contents)

        # Store contents
        if pass_by_reference:
            self._contents = contents
        else:
            self._contents = copy.deepcopy(contents)

        # Store whether file has a trailing newline
        if not isinstance(trailing_newline, bool):
            raise TypeError(
                'Argument "trailing_newline" must be of type "bool"')

        self.trailing_newline = trailing_newline

    def update_contents(self) -> None:
        """Updates the :py:attr:`contents` list based on object attributes

        This method by default does nothing.  However, it is intended that
        subclasses of :py:class:`TextFile` should override this method and
        define file-specific behavior in this method for converting custom
        object attributes to lines of text in the file, and storing these
        data in :py:attr:`contents`.

        For example, if defining a CSV-parser, the class might have an
        attribute that stores numerical data in a NumPy array, and the
        :py:meth:`update_contents` method might convert the data in this array
        to comma-separated strings and store them in :py:attr:`contents`.
        """
        return None

    def write(self, output_file: Union[str, pathlib.Path],
              write_mode: str = 'w', warn_before_overwrite: bool = True,
              prologue: str = '', epilogue: Optional[str] = None,
              line_ending: str = '\n', update_contents: bool = True,
              ) -> None:
        """Write file to disk

        Calling this method writes the file contents stored in
        :py:attr:`contents` to the disk.

        Parameters
        ----------
        output_file : str or pathlib.Path
            Output file to which to write content
        write_mode : str, optional
            Any mode (such as ``'w'`` or ``'a'``) for the built-in
            ``open()`` function for writing files (default is ``'w'``)
        warn_before_overwrite : bool, optional
            Whether to throw an error if ``output_file`` already exists
            (default is ``True``)
        prologue : str, optional
            Content written at beginning of file (default is ``''``)
        epilogue : str, optional
            Content written at end of file (default is to use the value of the
            ``line_ending`` argument if :py:attr:`trailing_newline` is
            ``True`` and ``''`` otherwise)
        line_ending : str, optional
            String written at the end of each line when writing file content
            (default is ``'\\n'``)
        update_contents : bool, optional
            Whether to call the :py:meth:`update_contents` method before
            writing the file (default is ``True``)
        """
        # Update contents from file attributes
        if update_contents:
            self.update_contents()

        # Confirm that "contents" attribute hasn't been modified improperly
        self._check_contents(self.contents)

        # Convert output file to `pathlib.Path`
        output_file = pathlib.Path(output_file).expanduser().resolve()

        # Check before overwriting file
        if warn_before_overwrite and output_file.is_file():
            raise FileExistsError(
                f'Output file "{output_file}" already exists')

        # Set epilogue
        if epilogue is None:
            epilogue = line_ending if self.trailing_newline else ''

        # Write output file
        with open(output_file, write_mode, encoding='utf_8', newline='') \
                as fileID:
            fileID.write(prologue)
            fileID.write(line_ending.join(self.contents))
            fileID.write(epilogue)



"""This module contains functions for parsing lists, tuples, and other
array-like data storage objects
"""

from .convert import convert_to_tuple
from .equality import is_array_equal
from .size import (
    check_len_equal,
    is_len_equal,
    max_list_item_len,
)



"""Customized exceptions for the :py:mod:`pyxx.files` module
"""


class NoFileSpecifiedError(AttributeError):
    """Error thrown if attempting to perform an operation that requires
    the ``pyxx.files.File.file`` attribute, but this attribute is set
    to ``None``"""

class UntrackedFileError(Exception):
    """Error thrown if trying to determine whether a file has been modified,
    but without having previously computed hashes of the file"""



"""This module provides a command-line interface (CLI) that can be used to
interact with the :py:class:`UnitConverterSI` class.
"""

import argparse
import difflib
import sys
from typing import List, Optional

# Importing the package version is a cylic import, but because the
# `__version__` variable is defined at the beginning of `pyxx/__init__.py`,
# the variable is cached and cyclic import does not cause problems
from pyxx import __version__ as PYXX_VERSION    # pylint: disable=R0401
from pyxx.units.classes.unitconverter import UnitConverterSI
from pyxx.units.exceptions import InvalidSearchFieldError


def execute_from_command_line(argv: Optional[List[str]] = None) -> None:
    """Calls the PyXX unit converter CLI

    Runs the PyXX unit converter CLI, allowing users to perform actions such
    as converting quantities from one unit to another or viewing information
    about units.

    Parameters
    ----------
    argv : list, optional
        List of arguments to pass to the PyXX unit converter CLI.  If not
        provided or ``None``, arguments are obtained from ``sys.argv``
        (default is ``None``)
    """
    if argv is None:
        argv = sys.argv[1:]

    sys.exit(UnitConverterCLI().execute(argv))


class UnitConverterCLI:
    """A command-line tool to interact with the PyXX unit conversion utilities

    This class is a wrapper around the :py:class:`UnitConverterSI` class,
    allowing users to access pre-defined units and perform basic searches and
    unit conversions from the terminal.  In most cases, to use this class,
    the :py:meth:`execute` method should be called with the desired arguments
    and flags in a :py:attr:`sys.argv`-style list.
    """

    def __init__(self) -> None:
        # Program metadata
        self.PROGRAM_NAME: str = 'unit-converter'

        self.PROGRAM_DESCRIPTION: str = '\n'.join([
            'PyXX unit converter command-line interface (CLI). This '
            'interface provides a\nsimple, terminal-based method to search '
            'and perform unit conversions with the\nPyXX package, without '
            'the need to set up scripts or launch a Python interpreter.\n'
            'For more information about unit conversions and available units,'
            'refer to the\nPyXX documentation (https://pyxx.readthedocs.io).',
            '',
            'Options:',
            '  -h, --help       Display this help message and exit',
            '      --version    Display PyXX unit converter version and exit',
            '',
            'Commands:',
            '  c, convert    Perform a unit conversion*',
            '  h, help       Display this help message and exit',
            '  i, info       Display information about a unit*',
            '  s, search     Search the units available in the unit converter*',
            '  v, version    Display PyXX unit converter version and exit',
            '',
            f'*Run \'{self.PROGRAM_NAME} COMMAND --help\' for more information '
            'about these commands',
        ]) + '\n'

        self.VERSION: str = PYXX_VERSION

        # Settings
        self._LINE_LENGTH: int = 80

    def help(self) -> int:
        """Prints the general help documentation for the CLI

        Outputs the general CLI usage information and command descriptions to
        the terminal.  More specific help (for each command) can be viewed by
        specifying the command with the "--help" flag.

        Returns
        -------
        int
            Exit code (0)
        """
        print(f'Usage:  {self.PROGRAM_NAME} [OPTIONS] COMMAND [COMMAND_OPTIONS]')
        print(f'\n{self.PROGRAM_DESCRIPTION}')

        return 0

    def execute(self, argv: List[str]) -> int:
        """Runs the PyXX unit converter CLI

        Parses a given set of :py:attr:`sys.argv`-style command-line arguments
        and displays to the terminal the results of the user-specified CLI
        actions.

        Parameters
        ----------
        argv : list
            A list of strings specifying the command-line arguments passed to
            the CLI.  Must be formatted in the same way as :py:attr:`sys.argv`

        Returns
        -------
        int
            Exit code

        Notes
        -----
        The first item in :py:attr:`sys.argv` is the script name, but this
        item should be omitted when calling :py:meth:`execute`.  In other
        words, :py:meth:`execute` should be passed an input argument ``argv``
        similar to ``sys.argv[1:]``.
        """
        try:
            command = argv[0]
            argv = argv[1:]
        except IndexError:
            return self._exit_with_error(message='No command provided',
                                         print_help=True)

        if command in ('convert', 'c'):
            return self._convert(argv)

        if command in ('help', 'h', '--help', '-h'):
            return self.help()

        if command in ('info', 'i'):
            return self._info(argv)

        if command in ('search', 's'):
            return self._search(argv)

        if command in ('version', 'v', '--version'):
            return self.version()

        return self._exit_with_error(message=f'Invalid command "{command}"',
                                     print_help=True)

    def version(self) -> int:
        """Displays the PyXX program version

        Returns
        -------
        int
            Exit code (0)
        """
        print(self.VERSION)

        return 0

    def _convert(self, argv: List[str]) -> int:
        # Process command-line arguments
        parser = argparse.ArgumentParser(
            prog=f'{self.PROGRAM_NAME} convert',
            description=(
                'Convert a quantity (either a single number or a comma-'
                'separated list of numbers) from one unit to another. Refer '
                'to the PyXX documentation (https://pyxx.readthedocs.io) or '
                'use the "search" command to view a list of available units.'
            ),
            add_help=False
        )

        required_args = parser.add_argument_group('Required arguments')
        optional_args = parser.add_argument_group('Optional arguments')

        required_args.add_argument(
            'quantity',
            action='store',
            type=str,
            help=('The quantity to be converted from one unit to another. '
                  'Either formatted as a number or a comma-separated list '
                  'of numbers')
        )

        required_args.add_argument(
            '-f', '--from',
            action='store',
            type=str,
            dest='from_unit',
            help='The original unit in which the quantity was expressed',
            required=True
        )
        required_args.add_argument(
            '-t', '--to',
            action='store',
            type=str,
            dest='to_unit',
            help='The unit to which the quantity is to be converted',
            required=True
        )

        optional_args.add_argument(
            '-h', '--help',
            action='help',
            help='Show this help message and exit'
        )

        args = parser.parse_args(argv)

        # Validate inputs
        from_unit = str(args.from_unit)
        to_unit = str(args.to_unit)

        for unit in (from_unit, to_unit):
            if not UnitConverterSI().is_defined_unit(unit):
                return self._exit_with_error(
                    f'Cannot perform unit conversion. Unit "{unit}" has not '
                    'been defined in the unit converter')

        if not UnitConverterSI().is_convertible(from_unit, to_unit):
            return self._exit_with_error(
                f'Cannot perform unit conversion. Units "{from_unit}" '
                f'and {to_unit} are not compatible')

        try:
            quantity = [float(x.strip()) for x in str(args.quantity).split(',')]
        except (TypeError, ValueError):
            return self._exit_with_error(
                f'Invalid format "{args.quantity}" of quantity to convert. '
                'Quantity must either be a number or a comma-separated list '
                'of numbers')

        # Perform unit conversion
        converted_quantity: List[float] = list(UnitConverterSI().convert(
            quantity  = quantity,
            from_unit = from_unit,
            to_unit   = to_unit
        ))

        print(','.join([str(x) for x in converted_quantity]))
        return 0

    def _exit_with_error(self, message: str, exit_code: int = 1,
                         print_help: bool = False) -> int:
        """Displays a CLI error message and returns a specified, nonzero
        exit code

        Parameters
        ----------
        message : str
            Error message to display
        exit_code : int, optional
            The exit code to return.  Must be a nonzero integer (default
            is ``1``)
        print_help : bool, optional
            Whether to print the CLI help (i.e., whether to call
            :py:meth:`help`) before displaying error message (default
            is ``False``)

        Returns
        -------
        int
            Exit code specified by ``exit_code``
        """
        if not isinstance(exit_code, int):
            raise TypeError('Exit code must be an integer')

        if not (0 < exit_code <= 255):
            raise ValueError('Exit code must be in the range (0, 255]')

        if print_help:
            self.help()
            print(f'{"-" * self._LINE_LENGTH}\n')

        print(f'{self.PROGRAM_NAME}: error: {message}')

        return exit_code

    def _info(self, argv: List[str]) -> int:
        # Process command-line arguments
        parser = argparse.ArgumentParser(
            prog=f'{self.PROGRAM_NAME} info',
            description=(
                'Display detailed information about a unit in the unit '
                'converter. Refer to the PyXX documentation '
                '(https://pyxx.readthedocs.io) or use the "search" command '
                'to view a list of available units.'
            ),
            add_help=False
        )

        required_args = parser.add_argument_group('Required arguments')
        optional_args = parser.add_argument_group('Optional arguments')

        required_args.add_argument(
            'unit',
            action='store',
            type=str,
            help='The unit about which to display detailed information'
        )

        optional_args.add_argument(
            '-h', '--help',
            action='help',
            help='Show this help message and exit'
        )

        args = parser.parse_args(argv)

        # Validate inputs
        unit = str(args.unit)

        if not UnitConverterSI().is_simplified_unit(unit):
            return self._exit_with_error(
                f'Unit "{unit}" is a compound unit. Detailed information can '
                'only be shown for simple units'
            )

        if not UnitConverterSI().is_defined_unit(unit):
            close_matches = difflib.get_close_matches(
                word=unit, possibilities=list(UnitConverterSI().keys()),
                n=5, cutoff=0.6)

            return self._exit_with_error(
                f'Unit "{unit}" has not been defined in the unit converter. '
                f'The most similar available units are: {close_matches}')

        # Display unit details
        unit_entry = UnitConverterSI()[unit]

        print(f'Unit ID:          {unit}')
        print(f'Name:             {unit_entry.name}')
        print(f'Description:      {unit_entry.description}')
        print(f'Tags:             {unit_entry.tags}')
        print(f'Aliases:          {UnitConverterSI().get_aliases(unit)}')
        print(f'Unit definition:  {unit_entry.unit}')

        return 0

    def _search(self, argv: List[str]) -> int:
        # Process command-line arguments
        parser = argparse.ArgumentParser(
            prog=f'{self.PROGRAM_NAME} search',
            description=(
                'Search the list of units and associated metadata for units '
                'defined in the unit converter. Refer to the PyXX '
                'documentation (https://pyxx.readthedocs.io) to view a list '
                'of available units.'
            ),
            add_help=False
        )

        required_args = parser.add_argument_group('Required arguments')
        optional_args = parser.add_argument_group('Optional arguments')

        required_args.add_argument(
            'search_term',
            action='store',
            type=str,
            help=('Search term. Use a wildcard (\'*\' or \'**\') to match '
                  'any string')
        )

        optional_args.add_argument(
            '-h', '--help',
            action='help',
            help='Show this help message and exit'
        )
        optional_args.add_argument(
            '--search-fields',
            action='store',
            type=str,
            default='key,name,tags,description',
            help=(
                'If provided, only the specified fields in the unit converter '
                'will be searched. Available search fields are: (\'key\', '
                '\'name\', \'tags\', \'description\'). Multiple fields can be '
                'provided as a comma-separated list. The default behavior is '
                'to search all available fields'
            )
        )
        optional_args.add_argument(
            '--filter-by-tags',
            action='store',
            type=str,
            default=None,
            help=(
                'If provided, only search results with the given tag(s) will '
                'be displayed. Use a comma-separated list to filter by '
                'multiple tags'
            )
        )
        optional_args.add_argument(
            '--hide-aliases',
            action='store_true',
            help=(
                'If this flag is provided, units with multiple aliases will '
                'be shown only once in the search results (only the first '
                'alias defined will be shown)'
            )
        )

        args = parser.parse_args(argv)

        if args.filter_by_tags is None:
            filter_by_tags = None
        else:
            filter_by_tags \
                = [x.strip() for x in str(args.filter_by_tags).split(',')]

        hide_aliases = bool(args.hide_aliases)
        search_fields = [x.strip() for x in str(args.search_fields).split(',')]
        search_term = str(args.search_term)

        # Display search results
        try:
            UnitConverterSI().search(
                search_term    = search_term,
                search_fields  = search_fields,
                filter_by_tags = filter_by_tags,
                hide_aliases   = hide_aliases,
                print_results  = True,
                return_results = False
            )
        except InvalidSearchFieldError as exception:
            return self._exit_with_error(str(exception))

        return 0



"""Functions for identifying the type of numerical data

The functions in this module can assist in identifying what type of data
is represented by a numerical type.
"""

from typing import Any


def is_float(value: Any):
    """Checks whether a value is a floating-point number

    Determines whether a value can be converted to a real number and
    expressed as a Python :py:class:`float`.

    Parameters
    ----------
    value : Any
        The value to be analyzed

    Returns
    -------
    bool
        Whether ``value`` represents a floating-point number

    Examples
    --------
    Decimal numbers and integers are identified as floating-point numbers:

    >>> pyxx.numbers.is_float(3.14)
    True
    >>> pyxx.numbers.is_float(3)
    True
    >>> pyxx.numbers.is_float(2/3)
    True

    This function checks if values *can be converted to* floating-point
    numbers, so string inputs may also be classified as valid floating-point
    numbers:

    >>> pyxx.numbers.is_float('-6.28')
    True
    """
    try:
        float(value)
    except (TypeError, ValueError):
        return False

    return True


def is_integer(value: Any):
    """Checks whether a value is an integer

    Parameters
    ----------
    value : Any
        The value to be analyzed

    Returns
    -------
    bool
        Whether ``value`` represents an integer

    Examples
    --------
    Integers (of type :py:class:`int` or :py:class:`float`) are identified
    as integers:

    >>> pyxx.numbers.is_integer(3)
    True
    >>> pyxx.numbers.is_integer(3.0)
    True
    >>> pyxx.numbers.is_integer(6/3.0)
    True

    This function checks if values *can be converted to* integers, so string
    inputs may also be classified as valid integers:

    >>> pyxx.numbers.is_integer('6.0')
    True

    However, floating-point numbers with a nonzero fractional part are not
    classified as valid integers:

    >>> pyxx.numbers.is_integer(0.5)
    False
    """
    return is_float(value) and float(value).is_integer()



"""This module contains classes that represent various types of files (binary
files, text files, etc.) and can be used to perform general file operations)
"""

from .binaryfile import BinaryFile
from .file import File
from .textfile import TextFile
